from sklearn.utils.estimator_checks import check_estimator

from dirty_cat import (
    DatetimeEncoder,
    GapEncoder,
    MinHashEncoder,
    SimilarityEncoder,
    TargetEncoder,
)


def test_sklearn_compatible_DatetimeEncoder():
    check_estimator(DatetimeEncoder())


def test_sklearn_compatible_GapEncoder():
    check_estimator(GapEncoder())


def test_sklearn_compatible_MinHashEncoder():
    check_estimator(MinHashEncoder())


def test_sklearn_compatible_SimilarityEncoder():
    check_estimator(SimilarityEncoder())


# def test_sklearn_compatible_SuperVectorizer():
#    check_estimator(SuperVectorizer())


def test_sklearn_compatible_TargetEncoder():
    check_estimator(TargetEncoder())



from pathlib import Path
from unittest import mock


@mock.patch("os.path.dirname")
def test_get_data_dir(mock_os_path_dirname):
    """
    Tests function ``get_data_dir()``.
    """
    from dirty_cat.datasets.utils import get_data_dir

    expected_return_value_default = Path("/user/directory/data")

    mock_os_path_dirname.return_value = "/user/directory/"
    assert get_data_dir() == expected_return_value_default

    expected_return_value_custom = expected_return_value_default / "tests"
    assert get_data_dir("tests") == expected_return_value_custom



"""
Fetching functions to retrieve example datasets, using
Scikit-Learn's ``fetch_openml()`` function.
"""

# Future notes:
# - Watch out for ``fetch_openml()`` API modifications:
# as of january 2021, the function is marked as experimental.


import gzip
import json
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Union

import pandas as pd
from sklearn.datasets import fetch_openml

from dirty_cat.datasets.utils import get_data_dir

# Directory where the ``.gz`` files containing the
# details on downloaded datasets are stored.
# Note: the tree structure is created by ``fetch_openml()``.
# As of october 2020, this function is annotated as
# ``Experimental`` so the structure might change in future releases.
# This path will be concatenated to the dirty_cat data directory,
# available via the function ``get_data_dir()``.
DETAILS_DIRECTORY: str = "openml/openml.org/api/v1/json/data/"

# Same as above; for the datasets features location.
FEATURES_DIRECTORY: str = "openml/openml.org/api/v1/json/data/features/"

# Same as above; for the datasets data location.
DATA_DIRECTORY: str = "openml/openml.org/data/v1/download/"

# The IDs of the datasets, from OpenML.
# For each dataset, its URL is constructed as follows:
openml_url: str = "https://www.openml.org/d/{ID}"
ROAD_SAFETY_ID: int = 42803
OPEN_PAYMENTS_ID: int = 42738
MIDWEST_SURVEY_ID: int = 42805
MEDICAL_CHARGE_ID: int = 42720
EMPLOYEE_SALARIES_ID: int = 42125
TRAFFIC_VIOLATIONS_ID: int = 42132
DRUG_DIRECTORY_ID: int = 43044


@dataclass(unsafe_hash=True)
class Details:
    name: str
    file_id: str
    description: str


@dataclass(unsafe_hash=True)
class Features:
    names: List[str]


@dataclass(unsafe_hash=True)
class DatasetAll:
    name: str
    description: str
    X: pd.DataFrame
    y: pd.Series
    source: str
    path: Path


@dataclass(unsafe_hash=True)
class DatasetInfoOnly:
    name: str
    description: str
    source: str
    target: str
    path: Path
    read_csv_kwargs: Dict[str, Any]


def fetch_openml_dataset(
    dataset_id: int,
    data_directory: Path = get_data_dir(),
) -> Dict[str, Any]:
    """
    Gets a dataset from OpenML (https://www.openml.org),
    or from the disk if already downloaded.

    Parameters
    ----------
    dataset_id: int
        The ID of the dataset to fetch.
    data_directory: Path
        Optional. A directory to save the data to.
        By default, the dirty_cat data directory.

    Returns
    -------
    Dict[str, Any]
        A dictionary containing:
          - ``description``: str
              The description of the dataset,
              as gathered from OpenML.
          - ``source``: str
              The dataset's URL from OpenML.
          - ``path``: pathlib.Path
              The local path leading to the dataset,
              saved as a CSV file.

    """
    # Make path absolute
    data_directory = data_directory.resolve()

    # Construct the path to the gzip file containing the details on a dataset.
    details_gz_path = data_directory / DETAILS_DIRECTORY / f"{dataset_id}.gz"
    features_gz_path = data_directory / FEATURES_DIRECTORY / f"{dataset_id}.gz"

    if not details_gz_path.is_file() or not features_gz_path.is_file():
        # If the details file or the features file don't exist,
        # download the dataset.
        warnings.warn(
            f"Could not find the dataset {dataset_id} locally. "
            "Downloading it from OpenML; this might take a while... "
            "If it is interrupted, some files might be invalid/incomplete: "
            "if on the following run, the fetching raises errors, you can try "
            f"fixing this issue by deleting the directory {data_directory!s}.",
            UserWarning,
            stacklevel=2,
        )
        _download_and_write_openml_dataset(
            dataset_id=dataset_id, data_directory=data_directory
        )
    details = _get_details(details_gz_path)

    # The file ID is required because the data file is named after this ID,
    # and not after the dataset's.
    file_id = details.file_id
    csv_path = data_directory / f"{details.name}.csv"

    data_gz_path = data_directory / DATA_DIRECTORY / f"{file_id}.gz"

    if not data_gz_path.is_file():
        # double-check.
        # If the data file does not exist, download the dataset.
        _download_and_write_openml_dataset(
            dataset_id=dataset_id, data_directory=data_directory
        )

    if not csv_path.is_file():
        # If the CSV file does not exist, use the dataset
        # downloaded by ``fetch_openml()`` to construct it.
        features = _get_features(features_gz_path)
        _export_gz_data_to_csv(data_gz_path, csv_path, features)

    url = openml_url.format(ID=dataset_id)

    return {
        "description": details.description,
        "source": url,
        "path": csv_path.resolve(),
    }


def _download_and_write_openml_dataset(dataset_id: int, data_directory: Path) -> None:
    """
    Downloads a dataset from OpenML,
    taking care of creating the directories.

    Parameters
    ----------
    dataset_id: int
        The ID of the dataset to download.
    data_directory: Path
        The directory in which the data will be saved.

    Raises
    ------
    ValueError
        If the ID is incorrect (does not exist on OpenML)
    urllib.error.URLError
        If there is no Internet connection.

    """
    # The ``fetch_openml()`` function returns a Scikit-Learn ``Bunch`` object,
    # which behaves just like a ``namedtuple``.
    # However, we do not want to save this data into memory:
    # we will read it from the disk later.
    fetch_openml(
        data_id=dataset_id,
        data_home=str(data_directory),
        as_frame=True,
    )


def _read_json_from_gz(compressed_dir_path: Path) -> dict:
    """
    Opens a gzip file, reads its content (JSON expected),
    and returns a dictionary.

    Parameters
    ----------
    compressed_dir_path: Path
        Path to the ``.gz`` file to read.

    Returns
    -------
    dict
        The information contained in the file,
        parsed from plain-text JSON.

    """
    if not compressed_dir_path.is_file():
        raise FileNotFoundError(f"Couldn't find file {compressed_dir_path!s}")

    # Read content
    with gzip.open(compressed_dir_path, mode="rt") as gz:
        content = gz.read()

    details_json = json.JSONDecoder().decode(content)
    return details_json


def _get_details(compressed_dir_path: Path) -> Details:
    """
    Gets useful details from the details file.

    Parameters
    ----------
    compressed_dir_path: Path
        The path to the ``.gz`` file containing the details.

    Returns
    -------
    Details
        A ``Details`` object.

    """
    details = _read_json_from_gz(compressed_dir_path)["data_set_description"]
    # We filter out the irrelevant information.
    # If you want to modify this list (to add or remove items)
    # you must also modify the ``Details`` object definition.
    return Details(
        name=details["name"],
        file_id=details["file_id"],
        description=details["description"],
    )


def _get_features(compressed_dir_path: Path) -> Features:
    """
    Gets features that can be inserted in the CSV file.
    The most important feature being the column names.

    Parameters
    ----------
    compressed_dir_path: Path
        Path to the gzip file containing the features.

    Returns
    -------
    Features
        A ``Features`` object.

    """
    raw_features = _read_json_from_gz(compressed_dir_path)["data_features"]
    # We filter out the irrelevant information.
    # If you want to modify this list (to add or remove items)
    # you must also modify the ``Features`` object definition.
    return Features(names=[column["name"] for column in raw_features["feature"]])


def _export_gz_data_to_csv(
    compressed_dir_path: Path, destination_file: Path, features: Features
) -> None:
    """
    Reads a gzip file containing ARFF data,
    and writes it to a target CSV.

    Parameters
    ----------
    compressed_dir_path: Path
        Path to the ``.gz`` file containing the ARFF data.
    destination_file: Path
        A CSV file to write to.
    features: Features
        A ``Features`` object containing the first CSV line (the column names).

    """
    atdata_found = False
    with destination_file.open(mode="w", encoding="utf8") as csv:
        with gzip.open(compressed_dir_path, mode="rt", encoding="utf8") as gz:
            csv.write(_features_to_csv_format(features))
            csv.write("\n")
            # We will look at each line of the file until we find
            # "@data": only after this tag is the actual CSV data.
            for line in gz.readlines():
                if not atdata_found:
                    if line.lower().startswith("@data"):
                        atdata_found = True
                else:
                    csv.write(line)


def _features_to_csv_format(features: Features) -> str:
    return ",".join(features.names)


def fetch_dataset_as_dataclass(
    dataset_name: str,
    dataset_id: int,
    target: str,
    read_csv_kwargs: dict,
    load_dataframe: bool,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """
    Takes a dataset identifier, a target column name,
    and some additional keyword arguments for `pd.read_csv`.

    If you don't need the dataset to be loaded in memory,
    pass `load_dataframe=False`.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`

    """
    info = fetch_openml_dataset(dataset_id)
    if load_dataframe:
        df = pd.read_csv(info["path"], **read_csv_kwargs)
        y = df[target]
        X = df.drop(target, axis="columns")
        dataset = DatasetAll(
            name=dataset_name,
            description=info["description"],
            X=X,
            y=y,
            source=info["source"],
            path=info["path"],
        )
    else:
        dataset = DatasetInfoOnly(
            name=dataset_name,
            description=info["description"],
            source=info["source"],
            target=target,
            path=info["path"],
            read_csv_kwargs=read_csv_kwargs,
        )

    return dataset


# Datasets fetchers section
# Public API


def fetch_employee_salaries(
    load_dataframe: bool = True,
    drop_linked: bool = True,
    drop_irrelevant: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the employee_salaries dataset (regression), available at
    https://openml.org/d/42125

    Description of the dataset:
    > Annual salary information including gross pay and overtime pay for all
    active, permanent employees of Montgomery County, MD paid in calendar
    year 2016. This information will be published annually each year.

    Parameters
    ----------
    drop_linked: bool (default True)
        Drops columns "2016_gross_pay_received" and "2016_overtime_pay",
        which are closely linked to "current_annual_salary", the target.

    drop_irrelevant: bool (default True)
        Drops column "full_name", which is usually irrelevant to the
        statistical analysis.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    dataset = fetch_dataset_as_dataclass(
        dataset_name="Employee salaries",
        dataset_id=EMPLOYEE_SALARIES_ID,
        target="current_annual_salary",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
            "na_values": ["?"],
        },
        load_dataframe=load_dataframe,
    )
    if load_dataframe:
        if drop_linked:
            dataset.X.drop(
                ["2016_gross_pay_received", "2016_overtime_pay"], axis=1, inplace=True
            )
        if drop_irrelevant:
            dataset.X.drop(["full_name"], axis=1, inplace=True)

    return dataset


def fetch_road_safety(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the road safety dataset (classification), available at
    https://openml.org/d/42803

    Description of the dataset:
    > Data reported to the police about the circumstances of personal injury
    road accidents in Great Britain from 1979, and the maker and model
    information of vehicles involved in the respective accident. This version
    includes data up to 2015.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Road safety",
        dataset_id=ROAD_SAFETY_ID,
        target="Sex_of_Driver",
        read_csv_kwargs={
            "na_values": ["?"],
        },
        load_dataframe=load_dataframe,
    )


def fetch_medical_charge(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the medical charge dataset (regression), available at
    https://openml.org/d/42720

    Description of the dataset:
    > The Inpatient Utilization and Payment Public Use File (Inpatient PUF)
    provides information on inpatient discharges for Medicare fee-for-service
    beneficiaries. The Inpatient PUF includes information on utilization,
    payment (total payment and Medicare payment), and hospital-specific charges
    for the more than 3,000 U.S. hospitals that receive Medicare Inpatient
    Prospective Payment System (IPPS) payments. The PUF is organized by
    hospital and Medicare Severity Diagnosis Related Group (MS-DRG) and
    covers Fiscal Year (FY) 2011 through FY 2016.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Medical charge",
        dataset_id=MEDICAL_CHARGE_ID,
        target="Average_Total_Payments",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
        },
        load_dataframe=load_dataframe,
    )


def fetch_midwest_survey(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the midwest survey dataset (classification), available at
    https://openml.org/d/42805

    Description of the dataset:
    > Survey to know if people self-identify as Midwesterners.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Midwest survey",
        dataset_id=MIDWEST_SURVEY_ID,
        target="Census_Region",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
        },
        load_dataframe=load_dataframe,
    )


def fetch_open_payments(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the open payments dataset (classification), available at
    https://openml.org/d/42738

    Description of the dataset:
    > Payments given by healthcare manufacturing companies to medical doctors
    or hospitals.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Open payments",
        dataset_id=OPEN_PAYMENTS_ID,
        target="status",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
            "na_values": ["?"],
        },
        load_dataframe=load_dataframe,
    )


def fetch_traffic_violations(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the traffic violations dataset (classification), available at
    https://openml.org/d/42132

    Description of the dataset:
    > This dataset contains traffic violation information from all electronic
    traffic violations issued in the Montgomery County, MD. Any information
    that can be used to uniquely identify the vehicle, the vehicle owner or
    the officer issuing the violation will not be published.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Traffic violations",
        dataset_id=TRAFFIC_VIOLATIONS_ID,
        target="violation_type",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
            "na_values": ["?"],
        },
        load_dataframe=load_dataframe,
    )


def fetch_drug_directory(
    load_dataframe: bool = True,
) -> Union[DatasetAll, DatasetInfoOnly]:
    """Fetches the drug directory dataset (classification), available at
    https://openml.org/d/43044

    Description of the dataset:
    > Product listing data submitted to the U.S. FDA for all unfinished,
    unapproved drugs.

    Returns
    -------
    DatasetAll
        If `load_dataframe=True`

    DatasetInfoOnly
        If `load_dataframe=False`
    """
    return fetch_dataset_as_dataclass(
        dataset_name="Drug directory",
        dataset_id=DRUG_DIRECTORY_ID,
        target="PRODUCTTYPENAME",
        read_csv_kwargs={
            "quotechar": "'",
            "escapechar": "\\",
        },
        load_dataframe=load_dataframe,
    )



import pytest

from dirty_cat.utils import LRUDict, Version


def test_lrudict():
    dict_ = LRUDict(10)

    for x in range(15):
        dict_[x] = f"filled {x}"

    for x in range(5, 15):
        assert x in dict_
        assert dict_[x] == f"filled {x}"

    for x in range(5):
        assert x not in dict_


def test_version():
    # Test those specified in its docstring
    assert Version("1.5") <= Version("1.6.5")
    assert Version("1.5") <= "1.6.5"

    assert (Version("1-5", separator="-") == Version("1-6-5", separator="-")) is False
    assert (Version("1-5", separator="-") == "1-6-5") is False
    with pytest.raises(ValueError):
        assert not (Version("1-5", separator="-") == "1.6.5")

    # Test all comparison methods
    assert Version("1.0") == Version("1.0")
    assert (Version("1.0") != Version("1.0")) is False

    assert Version("1.0") < Version("1.1")
    assert Version("1.1") <= Version("1.5")
    assert Version("1.1") <= Version("1.1")
    assert (Version("1.1") < Version("1.1")) is False

    assert Version("1.1") > Version("0.5")
    assert Version("1.1") >= Version("0.9")
    assert Version("1.1") >= Version("1.1")
    assert (Version("1.1") >= Version("1.9")) is False



import numpy as np
import pandas as pd
import pytest
from sklearn import __version__ as sklearn_version
from sklearn.datasets import fetch_20newsgroups

from dirty_cat import GapEncoder
from dirty_cat.utils import Version


def test_analyzer():
    """
    Test if the output is different when the analyzer is 'word' or 'char'.
    If it is, no error ir raised.
    """
    add_words = False
    n_samples = 70
    X_txt = fetch_20newsgroups(subset="train")["data"][:n_samples]
    X = np.array([X_txt, X_txt]).T
    n_components = 10
    # Test first analyzer output:
    encoder = GapEncoder(
        n_components=n_components,
        init="k-means++",
        analyzer="char",
        add_words=add_words,
        random_state=42,
        rescale_W=True,
    )
    encoder.fit(X)
    y = encoder.transform(X)

    # Test the other analyzer output:
    encoder = GapEncoder(
        n_components=n_components,
        init="k-means++",
        analyzer="word",
        add_words=add_words,
        random_state=42,
    )
    encoder.fit(X)
    y2 = encoder.transform(X)

    # Test inequality between the word and char analyzers output:
    np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, y, y2)


@pytest.mark.parametrize(
    "hashing, init, analyzer, add_words",
    [
        (False, "k-means++", "word", True),
        (True, "random", "char", False),
        (True, "k-means", "char_wb", True),
    ],
)
def test_gap_encoder(
    hashing: bool, init: str, analyzer: str, add_words: bool, n_samples: int = 70
) -> None:
    X_txt = fetch_20newsgroups(subset="train")["data"][:n_samples]
    X = np.array([X_txt, X_txt]).T
    n_components = 10
    # Test output shape
    encoder = GapEncoder(
        n_components=n_components,
        hashing=hashing,
        init=init,
        analyzer=analyzer,
        add_words=add_words,
        random_state=42,
        rescale_W=True,
    )
    encoder.fit(X)
    y = encoder.transform(X)
    assert y.shape == (n_samples, n_components * X.shape[1]), str(y.shape)

    # Test L1-norm of topics W.
    for col_enc in encoder.fitted_models_:
        l1_norm_W = np.abs(col_enc.W_).sum(axis=1)
        np.testing.assert_array_almost_equal(l1_norm_W, np.ones(n_components))

    # Test same seed return the same output
    encoder = GapEncoder(
        n_components=n_components,
        hashing=hashing,
        init=init,
        analyzer=analyzer,
        add_words=add_words,
        random_state=42,
    )
    encoder.fit(X)
    y2 = encoder.transform(X)
    np.testing.assert_array_equal(y, y2)


def test_input_type() -> None:
    # Numpy array with one column
    X = np.array([["alice"], ["bob"]])
    enc = GapEncoder(n_components=2, random_state=42)
    X_enc_array = enc.fit_transform(X)
    # List
    X2 = [["alice"], ["bob"]]
    enc = GapEncoder(n_components=2, random_state=42)
    X_enc_list = enc.fit_transform(X2)
    # Check if the encoded vectors are the same
    np.testing.assert_array_equal(X_enc_array, X_enc_list)

    # Numpy array with two columns
    X = np.array([["alice", "charlie"], ["bob", "delta"]])
    enc = GapEncoder(n_components=2, random_state=42)
    X_enc_array = enc.fit_transform(X)
    # Pandas dataframe with two columns
    df = pd.DataFrame(X)
    enc = GapEncoder(n_components=2, random_state=42)
    X_enc_df = enc.fit_transform(df)
    # Check if the encoded vectors are the same
    np.testing.assert_array_equal(X_enc_array, X_enc_df)


def test_partial_fit(n_samples=70) -> None:
    X_txt = fetch_20newsgroups(subset="train")["data"][:n_samples]
    X = np.array([X_txt, X_txt]).T
    # Gap encoder with fit on one batch
    enc = GapEncoder(random_state=42, batch_size=n_samples, max_iter=1)
    X_enc = enc.fit_transform(X)
    # Gap encoder with partial fit
    enc = GapEncoder(random_state=42)
    enc.partial_fit(X)
    X_enc_partial = enc.transform(X)
    # Check if the encoded vectors are the same
    np.testing.assert_almost_equal(X_enc, X_enc_partial)


def test_get_feature_names_out(n_samples=70) -> None:
    X_txt = fetch_20newsgroups(subset="train")["data"][:n_samples]
    X = np.array([X_txt, X_txt]).T
    enc = GapEncoder(random_state=42)
    enc.fit(X)
    # Expect a warning if sklearn >= 1.0
    if Version(sklearn_version) < "1.0":
        feature_names_1 = enc.get_feature_names()
    else:
        with pytest.warns(DeprecationWarning):
            feature_names_1 = enc.get_feature_names()
    feature_names_2 = enc.get_feature_names_out()
    for topic_labels in [feature_names_1, feature_names_2]:
        # Check number of labels
        assert len(topic_labels) == enc.n_components * X.shape[1]
        # Test different parameters for col_names
        topic_labels_2 = enc.get_feature_names_out(col_names="auto")
        assert topic_labels_2[0] == "col0: " + topic_labels[0]
        topic_labels_3 = enc.get_feature_names_out(col_names=["abc", "def"])
        assert topic_labels_3[0] == "abc: " + topic_labels[0]
    return


def test_overflow_error() -> None:
    np.seterr(over="raise", divide="raise")
    r = np.random.RandomState(0)
    X = r.randint(1e5, 1e6, size=(8000, 1)).astype(str)
    enc = GapEncoder(
        n_components=2, batch_size=1, min_iter=1, max_iter=1, random_state=0
    )
    enc.fit(X)


def test_score(n_samples: int = 70) -> None:
    X_txt = fetch_20newsgroups(subset="train")["data"][:n_samples]
    X1 = np.array(X_txt)[:, None]
    X2 = np.hstack([X1, X1])
    enc = GapEncoder(random_state=42)
    enc.fit(X1)
    score_X1 = enc.score(X1)
    enc.fit(X2)
    score_X2 = enc.score(X2)
    # Check that two identical columns give the same score
    assert score_X1 * 2 == score_X2


@pytest.mark.parametrize("missing", ["zero_impute", "error", "aaa"])
def test_missing_values(missing: str) -> None:
    observations = [
        ["alice", "bob"],
        ["bob", "alice"],
        ["bob", np.nan],
        ["alice", "charlie"],
        [np.nan, "alice"],
    ]
    observations = np.array(observations, dtype=object)
    enc = GapEncoder(handle_missing=missing, n_components=3)
    if missing == "error":
        with pytest.raises(ValueError, match="Input data contains missing values"):
            enc.fit_transform(observations)
    elif missing == "zero_impute":
        enc.fit_transform(observations)
        enc.partial_fit(observations)
    else:
        with pytest.raises(
            ValueError,
            match=r"handle_missing should be either "
            r"'error' or 'zero_impute', got 'aaa'",
        ):
            enc.fit_transform(observations)


if __name__ == "__main__":
    print("test_analyzer")
    test_analyzer()
    print("test_analyzer passed")

    print("start test_gap_encoder")
    test_gap_encoder(hashing=True, init="k-means++", analyzer="char", add_words=False)
    print("passed")

    print("start test_input_type")
    test_input_type()
    print("passed")

    print("start test_partial_fit")
    test_partial_fit()
    print("passed")

    print("start test_get_feature_names_out")
    test_get_feature_names_out()
    print("passed")

    print("start test_overflow_error")
    test_overflow_error()
    print("passed")

    print("start test_score")
    test_score()
    print("test_score passed")

    print("Done")



import collections
from typing import Dict, List, Literal, Union

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import check_array
from sklearn.utils.fixes import _object_dtype_isnan
from sklearn.utils.validation import check_is_fitted

from dirty_cat.utils import check_input


def lambda_(x, n):
    return x / (x + n)


class TargetEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical features as a numeric array given a target vector.

    Each category is encoded given the effect that it has in the
    target variable y. The method considers that categorical
    variables can present rare categories. It represents each category by the
    probability of y conditional on this category.
    In addition, it takes an empirical Bayes approach to shrink the estimate.


    Parameters
    ----------
    categories : typing.Union[typing.Literal["auto"], typing.List[typing.List[typing.Union[str, int]]]  # noqa
        Categories (unique values) per feature:
        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the i-th
          column. The passed categories must be sorted and should not mix
          strings and numeric values.
        The categories used can be found in the ``categories_`` attribute.
    clf_type : typing.Literal["regression", "binary-clf", "multiclass-clf"]
        The type of classification/regression problem.
    dtype : type, default=np.float64
        Desired dtype of output.
    handle_unknown : typing.Literal["error", "ignore"], default="error"
        Whether to raise an error or ignore if a unknown categorical feature is
        present during transform (default is to raise). When this parameter
        is set to 'ignore' and an unknown category is encountered during
        transform, the resulting one-hot encoded columns for this feature
        will be all zeros.
    handle_missing : typing.Literal["error", ""], default=""
        Whether to raise an error or impute with blank string '' if missing
        values (NaN) are present during fit (default is to impute).
        When this parameter is set to '', and a missing value is encountered
        during fit_transform, the resulting encoded columns for this feature
        will be all zeros.

    Attributes
    ----------
    n_features_in_: int
        Number of features in the data seen during fit.
    categories_ : typing.List[np.array]
        The categories of each feature determined during fitting
        (in order corresponding with output of ``transform``).

    References
    -----------
    For more details, see Micci-Barreca, 2001: A preprocessing scheme for
    high-cardinality categorical attributes in classification and prediction
    problems.
    """

    n_features_in_: int
    _label_encoders_: List[LabelEncoder]
    categories_: List[np.array]
    n_: int

    def __init__(
        self,
        categories: Union[Literal["auto"], List[Union[List[str], np.array]]] = "auto",
        clf_type: Literal["regression", "binary-clf", "multiclass-clf"] = "binary-clf",
        dtype: type = np.float64,
        handle_unknown: Literal["error", "ignore"] = "error",
        handle_missing: Literal["error", ""] = "",
    ):
        self.categories = categories
        self.dtype = dtype
        self.clf_type = clf_type
        self.handle_unknown = handle_unknown
        self.handle_missing = handle_missing

    def _more_tags(self) -> Dict[str, List[str]]:
        """
        Used internally by sklearn to ease the estimator checks.
        """
        return {"X_types": ["categorical"]}

    def fit(self, X, y) -> "TargetEncoder":
        """
        Fit the TargetEncoder to X.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to determine the categories of each feature.
        y : array
            The associated target vector.

        Returns
        -------
        TargetEncoder
            Fitted TargetEncoder instance.
        """
        X = check_input(X)
        self.n_features_in_ = X.shape[1]
        if self.handle_missing not in ["error", ""]:
            raise ValueError(
                f"Got handle_missing={self.handle_missing!r}, but expected "
                "any of {'error', ''}. "
            )
        if hasattr(X, "iloc") and X.isna().values.any():
            if self.handle_missing == "error":
                raise ValueError(
                    "Found missing values in input data; set "
                    "handle_missing='' to encode with missing values. "
                )
            else:
                X = X.fillna(self.handle_missing)
        elif not hasattr(X, "dtype") and isinstance(X, list):
            X = np.asarray(X, dtype=object)

        if hasattr(X, "dtype"):
            mask = _object_dtype_isnan(X)
            if X.dtype.kind == "O" and mask.any():
                if self.handle_missing == "error":
                    raise ValueError(
                        "Found missing values in input data; set "
                        "handle_missing='' to encode with missing values. "
                    )
                else:
                    X[mask] = self.handle_missing

        if self.handle_unknown not in ["error", "ignore"]:
            raise ValueError(
                f"Got handle_unknown={self.handle_unknown!r}, but expected"
                'any of {"error", "ignore"}. '
            )

        if self.categories != "auto":
            for cats in self.categories:
                if not np.all(np.sort(cats) == np.array(cats)):
                    raise ValueError("Unsorted categories are not yet supported. ")

        X_temp = check_array(X, dtype=None)
        if not hasattr(X, "dtype") and np.issubdtype(X_temp.dtype, np.str_):
            X = check_array(X, dtype=np.object)
        else:
            X = X_temp
        X: np.array

        n_samples, n_features = X.shape

        self._label_encoders_ = [LabelEncoder() for _ in range(n_features)]

        for j in range(n_features):
            le = self._label_encoders_[j]
            Xj = X[:, j]
            if self.categories == "auto":
                le.fit(Xj)
            else:
                if self.handle_unknown == "error":
                    valid_mask = np.in1d(Xj, self.categories[j])
                    if not np.all(valid_mask):
                        diff = np.unique(Xj[~valid_mask])
                        raise ValueError(
                            f"Found unknown categories {diff} in column {j} during fit"
                        )
                le.classes_ = np.array(self.categories[j])

        self.categories_ = [le.classes_ for le in self._label_encoders_]
        self.n_ = len(y)
        if self.clf_type in ["binary-clf", "regression"]:
            self.Eyx_ = [
                {cat: np.mean(y[X[:, j] == cat]) for cat in self.categories_[j]}
                for j in range(len(self.categories_))
            ]
            self.Ey_ = np.mean(y)
            self.counter_ = {j: collections.Counter(X[:, j]) for j in range(n_features)}
        if self.clf_type in ["multiclass-clf"]:
            self.classes_ = np.unique(y)

            self.Eyx_ = {
                c: [
                    {
                        cat: np.mean((y == c)[X[:, j] == cat])
                        for cat in self.categories_[j]
                    }
                    for j in range(len(self.categories_))
                ]
                for c in self.classes_
            }
            self.Ey_ = {c: np.mean(y == c) for c in self.classes_}
            self.counter_ = {j: collections.Counter(X[:, j]) for j in range(n_features)}
        self.k_ = {j: len(self.counter_[j]) for j in self.counter_}
        return self

    def transform(self, X) -> np.array:
        """
        Transform X using the specified encoding scheme.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features_new]
            The data to encode.

        Returns
        -------
        2-d np.array
            Transformed input.
        """
        check_is_fitted(self, attributes=["n_features_in_"])
        X = check_input(X)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                f"The number of features in the input data ({X.shape[1]}) "
                "does not match the number of features "
                f"seen during fit ({self.n_features_in_})."
            )
        if hasattr(X, "iloc") and X.isna().values.any():
            if self.handle_missing == "error":
                raise ValueError(
                    "Found missing values in input data; set "
                    "handle_missing='' to encode with missing values. "
                )
            if self.handle_missing != "error":
                X = X.fillna(self.handle_missing)
        elif not hasattr(X, "dtype") and isinstance(X, list):
            X = np.asarray(X, dtype=object)

        if hasattr(X, "dtype"):
            mask = _object_dtype_isnan(X)
            if X.dtype.kind == "O" and mask.any():
                if self.handle_missing == "error":
                    raise ValueError(
                        "Found missing values in input data; set "
                        "handle_missing='' to encode with missing values. "
                    )
                if self.handle_missing != "error":
                    X[mask] = self.handle_missing

        X_temp = check_array(X, dtype=None)
        if not hasattr(X, "dtype") and np.issubdtype(X_temp.dtype, np.str_):
            X = check_array(X, dtype=np.object)
        else:
            X = X_temp

        n_samples, n_features = X.shape
        X_int = np.zeros_like(X, dtype=int)
        X_mask = np.ones_like(X, dtype=bool)

        for i in range(n_features):
            Xi = X[:, i]
            valid_mask = np.in1d(Xi, self.categories_[i])

            if not np.all(valid_mask):
                if self.handle_unknown == "error":
                    diff = np.unique(X[~valid_mask, i])
                    raise ValueError(
                        f"Found unknown categories {diff} in column {i} "
                        "during transform. "
                    )
                else:
                    # Set the problematic rows to an acceptable value and
                    # continue `The rows are marked `X_mask` and will be
                    # removed later.
                    X_mask[:, i] = valid_mask
                    Xi = Xi.copy()
                    Xi[~valid_mask] = self.categories_[i][0]
            X_int[:, i] = self._label_encoders_[i].transform(Xi)

        out = []

        for j, cats in enumerate(self.categories_):
            unqX = np.unique(X[:, j])
            encoder = {x: 0 for x in unqX}
            if self.clf_type in ["binary-clf", "regression"]:
                for x in unqX:
                    if x not in cats:
                        Eyx = 0
                    else:
                        Eyx = self.Eyx_[j][x]
                    lambda_n = lambda_(self.counter_[j][x], self.n_ / self.k_[j])
                    encoder[x] = lambda_n * Eyx + (1 - lambda_n) * self.Ey_
                x_out = np.zeros((len(X[:, j]), 1))
                for i, x in enumerate(X[:, j]):
                    x_out[i, 0] = encoder[x]
                out.append(x_out.reshape(-1, 1))
            if self.clf_type == "multiclass-clf":
                x_out = np.zeros((len(X[:, j]), len(self.classes_)))
                lambda_n = {x: 0 for x in unqX}
                for x in unqX:
                    lambda_n[x] = lambda_(self.counter_[j][x], self.n_ / self.k_[j])
                for k, c in enumerate(np.unique(self.classes_)):
                    for x in unqX:
                        if x not in cats:
                            Eyx = 0
                        else:
                            Eyx = self.Eyx_[c][j][x]
                        encoder[x] = lambda_n[x] * Eyx + (1 - lambda_n[x]) * self.Ey_[c]
                    for i, x in enumerate(X[:, j]):
                        x_out[i, k] = encoder[x]
                out.append(x_out)
        out = np.hstack(out)
        return out



import random
import time
from string import ascii_lowercase

import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import fetch_20newsgroups

from dirty_cat import MinHashEncoder
from dirty_cat.datasets import fetch_employee_salaries


def test_MinHashEncoder(n_sample: int = 70) -> None:
    X_txt = fetch_20newsgroups(subset="train")["data"]
    X = np.array(X_txt[:n_sample])[:, None]

    for minmax_hash in [True, False]:
        for hashing in ["fast", "murmur"]:
            if minmax_hash and hashing == "murmur":
                pass  # not implemented

            # Test output shape
            encoder = MinHashEncoder(n_components=50, hashing=hashing)
            encoder.fit(X)
            y = encoder.transform(X)
            assert y.shape == (n_sample, 50), str(y.shape)
            assert len(set(y[0])) == 50

            # Test same seed return the same output
            encoder = MinHashEncoder(50, hashing=hashing)
            encoder.fit(X)
            y2 = encoder.transform(X)
            np.testing.assert_array_equal(y, y2)

            # Test min property
            if not minmax_hash:
                X_substring = [x[: x.find(" ")] for x in X[:, 0]]
                X_substring = np.array(X_substring)[:, None]
                encoder = MinHashEncoder(50, hashing=hashing)
                encoder.fit(X_substring)
                y_substring = encoder.transform(X_substring)
                np.testing.assert_array_less(y - y_substring, 0.0001)


def test_multiple_columns() -> None:
    """
    This test aims at verifying that fitting multiple columns
    with the MinHashEncoder will not produce an error, and will
    encode the column independently.
    """
    X = pd.DataFrame(
        [
            ("bird", "parrot"),
            ("bird", "nightingale"),
            ("mammal", "monkey"),
            ("mammal", np.nan),
        ],
        columns=("class", "type"),
    )
    X1 = X[["class"]]
    X2 = X[["type"]]
    fit1 = MinHashEncoder(n_components=30).fit_transform(X1)
    fit2 = MinHashEncoder(n_components=30).fit_transform(X2)
    fit = MinHashEncoder(n_components=30).fit_transform(X)
    assert np.array_equal(
        np.array([fit[:, :30], fit[:, 30:60]]), np.array([fit1, fit2])
    )


def test_input_type() -> None:
    # Numpy array
    X = np.array(["alice", "bob"])[:, None]
    enc = MinHashEncoder(n_components=2)
    enc.fit_transform(X)
    # List
    X = [["alice"], ["bob"]]
    enc = MinHashEncoder(n_components=2)
    enc.fit_transform(X)


def profile_encoder(hashing: str = "fast", minmax_hash: bool = False) -> float:
    # not a unit test
    employee_salaries = fetch_employee_salaries()
    df = employee_salaries.X
    X = df[["employee_position_title"]]
    t0 = time.time()
    enc = MinHashEncoder(n_components=50, hashing=hashing, minmax_hash=minmax_hash)
    enc.fit(X)
    y = enc.transform(X)
    assert y.shape == (len(X), 50)
    eta = time.time() - t0
    return eta


@pytest.mark.parametrize(
    "input_type, missing, hashing",
    [
        ["numpy", "error", "fast"],
        ["pandas", "zero_impute", "murmur"],
        ["numpy", "zero_impute", "fast"],
    ],
)
def test_missing_values(input_type: str, missing: str, hashing: str) -> None:
    X = ["Red", np.nan, "green", "blue", "green", "green", "blue", float("nan")]
    n = 3
    z = np.zeros(n)

    if input_type == "numpy":
        X = np.array(X, dtype=object)[:, None]
    elif input_type == "pandas":
        X = pd.DataFrame(X)

    encoder = MinHashEncoder(
        n_components=n, hashing=hashing, minmax_hash=False, handle_missing=missing
    )
    if missing == "error":
        encoder.fit(X)
        if input_type in ["numpy", "pandas"]:
            with pytest.raises(ValueError, match=r"missing values in input"):
                encoder.transform(X)
    elif missing == "zero_impute":
        encoder.fit(X)
        y = encoder.transform(X)
        if input_type == "list":
            assert np.allclose(y[1], y[-1])
        else:
            assert np.array_equal(y[1], z)
            assert np.array_equal(y[-1], z)
    else:
        with pytest.raises(ValueError, match=r"expected any of"):
            encoder.fit_transform(X)
    return


def test_cache_overflow() -> None:
    # Regression test for cache overflow resulting in -1s in encoding
    def get_random_string(length):
        letters = ascii_lowercase
        result_str = "".join(random.choice(letters) for _ in range(length))
        return result_str

    encoder = MinHashEncoder(n_components=3)
    capacity = encoder._capacity
    raw_data = [get_random_string(10) for _ in range(capacity + 1)]
    raw_data = np.array(raw_data)[:, None]
    y = encoder.fit_transform(raw_data)

    assert len(y[y == -1.0]) == 0


if __name__ == "__main__":
    print("start test_MinHashEncoder")
    test_MinHashEncoder()
    print("passed")

    print("start test_multiple_columns")
    test_multiple_columns()
    print("passed")

    for _ in range(3):
        print("time profile_encoder(MinHashEncoder, hashing=fast)")
        print(f"{profile_encoder(hashing='fast'):.4} seconds")

    for _ in range(3):
        print("time profile_encoder(MinHashEncoder, hashing=fast) with minmax")
        print(f"{profile_encoder(hashing='fast', minmax_hash=True):.4} seconds")

    print("time profile_encoder(MinHashEncoder, hashing=murmur)")
    print(f"{profile_encoder(hashing='murmur'):.4} seconds")

    print("Done")



from typing import Callable, Optional

import numpy as np
import numpy.testing
import pytest
from sklearn import __version__ as sklearn_version

from dirty_cat import similarity_encoder, string_distances
from dirty_cat.similarity_encoder import get_kmeans_prototypes
from dirty_cat.utils import Version


def test_specifying_categories() -> None:
    # When creating a new SimilarityEncoder:
    # - if categories = 'auto', the categories are the sorted, unique training
    # set observations (for each column)
    # - if categories is a list (of lists), the categories for each column are
    # each item in the list

    # In this test, we first find the sorted, unique categories in the training
    # set, and create a SimilarityEncoder by giving it explicitly the computed
    # categories. The test consists in making sure the transformed observations
    # given by this encoder are equal to the transformed observations in the
    # case of a SimilarityEncoder created with categories = 'auto'

    observations = [["bar"], ["foo"]]
    categories = [["bar", "foo"]]

    sim_enc_with_cat = similarity_encoder.SimilarityEncoder(
        categories=categories, ngram_range=(2, 3)
    )
    sim_enc_auto_cat = similarity_encoder.SimilarityEncoder(ngram_range=(2, 3))

    feature_matrix_with_cat = sim_enc_with_cat.fit_transform(observations)
    feature_matrix_auto_cat = sim_enc_auto_cat.fit_transform(observations)

    assert np.allclose(feature_matrix_auto_cat, feature_matrix_with_cat)


def test_fast_ngram_similarity() -> None:
    vocabulary = [["bar", "foo"]]
    observations = [["foo"], ["baz"]]

    sim_enc = similarity_encoder.SimilarityEncoder(
        ngram_range=(2, 2), categories=vocabulary
    )

    sim_enc.fit(observations)
    feature_matrix = sim_enc.transform(observations, fast=False)
    feature_matrix_fast = sim_enc.transform(observations, fast=True)

    assert np.allclose(feature_matrix, feature_matrix_fast)


def _test_missing_values(input_type, missing):
    observations = [["a", "b"], ["b", "a"], ["b", np.nan], ["a", "c"], [np.nan, "a"]]
    encoded = np.array(
        [
            [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
        ]
    )

    if input_type == "numpy":
        observations = np.array(observations, dtype=object)
    elif input_type == "pandas":
        pd = pytest.importorskip("pandas")
        observations = pd.DataFrame(observations)

    sim_enc = similarity_encoder.SimilarityEncoder(handle_missing=missing)
    if missing == "error":
        with pytest.raises(ValueError, match=r"Found missing values in input"):
            sim_enc.fit_transform(observations)
    elif missing == "":
        ans = sim_enc.fit_transform(observations)
        assert np.allclose(encoded, ans)
    else:
        with pytest.raises(ValueError, match=r"expected any of"):
            sim_enc.fit_transform(observations)
        return


def _test_missing_values_transform(input_type: str, missing: str) -> None:
    observations = [["a", "b"], ["b", "a"], ["b", "b"], ["a", "c"], ["c", "a"]]
    test_observations = [
        ["a", "b"],
        ["b", "a"],
        ["b", np.nan],
        ["a", "c"],
        [np.nan, "a"],
    ]
    encoded = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
        ]
    )

    if input_type == "numpy":
        test_observations = np.array(test_observations, dtype=object)
    elif input_type == "pandas":
        pd = pytest.importorskip("pandas")
        test_observations = pd.DataFrame(test_observations)

    sim_enc = similarity_encoder.SimilarityEncoder(handle_missing=missing)
    if missing == "error":
        sim_enc.fit_transform(observations)
        with pytest.raises(ValueError, match=r"Found missing values in input"):
            sim_enc.transform(test_observations)
    elif missing == "":
        sim_enc.fit_transform(observations)
        ans = sim_enc.transform(test_observations)
        assert np.allclose(encoded, ans)


def _test_similarity(
    similarity_f: Callable,
    hashing_dim: Optional[int] = None,
    categories: str = "auto",
    n_prototypes: int = None,
) -> None:
    if n_prototypes is None:
        X = np.array(["aa", "aaa", "aaab"]).reshape(-1, 1)
        X_test = np.array([["Aa", "aAa", "aaa", "aaab", " aaa  c"]]).reshape(-1, 1)

        model = similarity_encoder.SimilarityEncoder(
            hashing_dim=hashing_dim,
            categories=categories,
            n_prototypes=n_prototypes,
            ngram_range=(3, 3),
        )

        encoder = model.fit(X).transform(X_test)

        ans = np.zeros((len(X_test), len(X)))
        for i, x_t in enumerate(X_test.reshape(-1)):
            for j, x in enumerate(X.reshape(-1)):
                ans[i, j] = similarity_f(x_t, x, 3)
        numpy.testing.assert_almost_equal(encoder, ans)
    else:
        X = np.array(
            ["aac", "aaa", "aaab", "aaa", "aaab", "aaa", "aaab", "aaa"]
        ).reshape(-1, 1)
        X_test = np.array([["Aa", "aAa", "aaa", "aaab", " aaa  c"]]).reshape(-1, 1)

        try:
            model = similarity_encoder.SimilarityEncoder(
                hashing_dim=hashing_dim,
                categories=categories,
                n_prototypes=n_prototypes,
                random_state=42,
                ngram_range=(3, 3),
            )
        except ValueError as e:
            assert (
                e.__str__()
                == "n_prototypes expected None or a positive non null integer"
            )
            return

        encoder = model.fit(X).transform(X_test)
        if n_prototypes == 1:
            assert model.categories_ == ["aaa"]
        elif n_prototypes == 2:
            a = [np.array(["aaa", "aaab"], dtype="<U4")]
            assert np.array_equal(a, model.categories_)
        elif n_prototypes == 3:
            a = [np.array(["aaa", "aaab", "aac"], dtype="<U4")]
            assert np.array_equal(a, model.categories_)

        ans = np.zeros((len(X_test), len(np.array(model.categories_).reshape(-1))))
        for i, x_t in enumerate(X_test.reshape(-1)):
            for j, x in enumerate(np.array(model.categories_).reshape(-1)):
                ans[i, j] = similarity_f(x_t, x, 3)

        numpy.testing.assert_almost_equal(encoder, ans)


def test_similarity_encoder() -> None:
    categories = ["auto", "most_frequent", "k-means"]
    for category in categories:
        if category == "auto":
            _test_similarity(
                string_distances.ngram_similarity,
                categories=category,
                n_prototypes=None,
            )
            _test_similarity(
                string_distances.ngram_similarity,
                hashing_dim=2**16,
                categories=category,
            )
        else:
            for i in range(1, 4):
                _test_similarity(
                    string_distances.ngram_similarity,
                    categories=category,
                    n_prototypes=i,
                )
                _test_similarity(
                    string_distances.ngram_similarity,
                    hashing_dim=2**16,
                    categories=category,
                    n_prototypes=i,
                )

    input_types = ["list", "numpy", "pandas"]
    handle_missing = ["aaa", "error", ""]
    for input_type in input_types:
        for missing in handle_missing:
            _test_missing_values(input_type, missing)
            _test_missing_values_transform(input_type, missing)


def test_kmeans_protoypes() -> None:
    X_test = np.array(["cbbba", "baaac", "accc"])
    proto = get_kmeans_prototypes(X_test, 3)
    assert np.array_equal(np.sort(proto), np.sort(X_test))


def test_reproducibility() -> None:
    sim_enc = similarity_encoder.SimilarityEncoder(
        categories="k-means",
        n_prototypes=10,
        random_state=435,
    )
    X = np.array([" %s " % chr(i) for i in range(32, 127)]).reshape((-1, 1))
    prototypes = sim_enc.fit(X).categories_[0]
    for i in range(10):
        assert np.array_equal(prototypes, sim_enc.fit(X).categories_[0])


def test_get_features() -> None:
    # See https://github.com/dirty-cat/dirty_cat/issues/168
    sim_enc = similarity_encoder.SimilarityEncoder(random_state=435)
    X = np.array(["%s" % chr(i) for i in range(32, 127)]).reshape((-1, 1))
    sim_enc.fit(X)
    if Version(sklearn_version) < Version("1.0"):
        feature_names = sim_enc.get_feature_names()
    else:
        feature_names = sim_enc.get_feature_names_out()
    assert feature_names.tolist() == [
        "x0_ ",
        "x0_!",
        'x0_"',
        "x0_#",
        "x0_$",
        "x0_%",
        "x0_&",
        "x0_'",
        "x0_(",
        "x0_)",
        "x0_*",
        "x0_+",
        "x0_,",
        "x0_-",
        "x0_.",
        "x0_/",
        "x0_0",
        "x0_1",
        "x0_2",
        "x0_3",
        "x0_4",
        "x0_5",
        "x0_6",
        "x0_7",
        "x0_8",
        "x0_9",
        "x0_:",
        "x0_;",
        "x0_<",
        "x0_=",
        "x0_>",
        "x0_?",
        "x0_@",
        "x0_A",
        "x0_B",
        "x0_C",
        "x0_D",
        "x0_E",
        "x0_F",
        "x0_G",
        "x0_H",
        "x0_I",
        "x0_J",
        "x0_K",
        "x0_L",
        "x0_M",
        "x0_N",
        "x0_O",
        "x0_P",
        "x0_Q",
        "x0_R",
        "x0_S",
        "x0_T",
        "x0_U",
        "x0_V",
        "x0_W",
        "x0_X",
        "x0_Y",
        "x0_Z",
        "x0_[",
        "x0_\\",
        "x0_]",
        "x0_^",
        "x0__",
        "x0_`",
        "x0_a",
        "x0_b",
        "x0_c",
        "x0_d",
        "x0_e",
        "x0_f",
        "x0_g",
        "x0_h",
        "x0_i",
        "x0_j",
        "x0_k",
        "x0_l",
        "x0_m",
        "x0_n",
        "x0_o",
        "x0_p",
        "x0_q",
        "x0_r",
        "x0_s",
        "x0_t",
        "x0_u",
        "x0_v",
        "x0_w",
        "x0_x",
        "x0_y",
        "x0_z",
        "x0_{",
        "x0_|",
        "x0_}",
        "x0_~",
    ]



"""
dirty_cat: Learning on dirty categories.
"""
import os

from .datetime_encoder import DatetimeEncoder
from .gap_encoder import GapEncoder
from .minhash_encoder import MinHashEncoder
from .similarity_encoder import SimilarityEncoder
from .super_vectorizer import SuperVectorizer
from .target_encoder import TargetEncoder

version_file = os.path.join(os.path.dirname(__file__), "VERSION.txt")
with open(version_file) as fh:
    __version__ = fh.read().strip()

__all__ = [
    "SimilarityEncoder",
    "TargetEncoder",
    "MinHashEncoder",
    "GapEncoder",
    "DatetimeEncoder",
    "SuperVectorizer",
]



# Inspired from :
# https://github.com/scikit-learn/scikit-learn/blob/main/benchmarks/bench_text_vectorizers.py
# plot run time and memory usage of the different encoders using the SuperVectorizer method
# test if auto_cast=True plays an important role

"""

To run this benchmark, you will need,

 * dirty_cat
 * pandas
 * memory_profiler

"""
import timeit
import itertools

import numpy as np
import pandas as pd
from memory_profiler import memory_usage

from dirty_cat.datasets import fetch_traffic_violations

from dirty_cat import SimilarityEncoder, GapEncoder, MinHashEncoder
from dirty_cat import SuperVectorizer

n_repeat = 3


def run_encoders(Encoder, X, **params):
    def f():
        enc = Encoder(**params)
        enc.fit_transform(X)

    return f


data = fetch_traffic_violations()

X = data[1][:5000]
y = data[2][:5000]

print("=" * 80 + "\n#" + "    Dirty_cat encoders benchmark" + "\n" + "=" * 80 + "\n")
print(f"Using a subset of the traffic violations dataset ({len(X)} observations).")
print("This benchmarks runs in ~1 min ...")

res = []

for Encoder, (auto_cast, high_card_cat_transformer) in itertools.product(
    [SuperVectorizer],
    [
        (True, SimilarityEncoder()),
        (True, GapEncoder()),
        (True, MinHashEncoder()),
        (False, SimilarityEncoder()),
        (False, GapEncoder()),
        (False, MinHashEncoder()),
    ],
):
    params = {
        "auto_cast": auto_cast,
        "high_card_cat_transformer": high_card_cat_transformer,
    }
    bench = {"encoder": Encoder}
    bench.update(params)
    dt = timeit.repeat(run_encoders(Encoder, X, **params), number=1, repeat=n_repeat)
    bench["time"] = f"{np.mean(dt):.3f} (+-{np.std(dt):.3f})"

    mem_usage = memory_usage(run_encoders(Encoder, X, **params))

    bench["memory"] = f"{np.max(mem_usage):.1f}"

    res.append(bench)

df = pd.DataFrame(res).set_index(["auto_cast", "high_card_cat_transformer", "encoder"])

print("\n========== Run time performance (sec) ===========\n")
print(
    "Computing the mean and the standard deviation "
    f"of the run time over {n_repeat} runs...\n"
)
print(df["time"].unstack(level=-1))

print("\n=============== Memory usage (MB) ===============\n")
print(df["memory"].unstack(level=-1))



"""
This test suite ensures the docstrings of class methods in
dirty_cat are formatted according to numpydoc specifications.

`FUNCTION_DOCSTRING_IGNORE_SET` defines a set of class methods
to skip while running the validation tests, so that CI will
not fail.

Therefore, developpers having formatted methods to numpydoc
should also remove their corresponding references from the list.
"""
import inspect

import re
from importlib import import_module
from typing import Optional

import pytest

numpydoc_validation = pytest.importorskip("numpydoc.validate")

FUNCTION_DOCSTRING_IGNORE_SET = {
    "dirty_cat.datetime_encoder.DatetimeEncoder",
    "dirty_cat.datetime_encoder.DatetimeEncoder.fit",
    "dirty_cat.datetime_encoder.DatetimeEncoder.fit_transform",
    "dirty_cat.datetime_encoder.DatetimeEncoder.get_feature_names",
    "dirty_cat.datetime_encoder.DatetimeEncoder.get_feature_names_out",
    "dirty_cat.datetime_encoder.DatetimeEncoder.get_params",
    "dirty_cat.datetime_encoder.DatetimeEncoder.set_params",
    "dirty_cat.datetime_encoder.DatetimeEncoder.transform",
    "dirty_cat.gap_encoder.GapEncoder",
    "dirty_cat.gap_encoder.GapEncoder.fit",
    "dirty_cat.gap_encoder.GapEncoder.fit_transform",
    "dirty_cat.gap_encoder.GapEncoder.get_feature_names",
    "dirty_cat.gap_encoder.GapEncoder.get_feature_names_out",
    "dirty_cat.gap_encoder.GapEncoder.get_params",
    "dirty_cat.gap_encoder.GapEncoder.partial_fit",
    "dirty_cat.gap_encoder.GapEncoder.score",
    "dirty_cat.gap_encoder.GapEncoder.set_params",
    "dirty_cat.gap_encoder.GapEncoder.transform",
    "dirty_cat.minhash_encoder.MinHashEncoder",
    "dirty_cat.minhash_encoder.MinHashEncoder.fit",
    "dirty_cat.minhash_encoder.MinHashEncoder.fit_transform",
    "dirty_cat.minhash_encoder.MinHashEncoder.get_fast_hash",
    "dirty_cat.minhash_encoder.MinHashEncoder.get_params",
    "dirty_cat.minhash_encoder.MinHashEncoder.get_unique_ngrams",
    "dirty_cat.minhash_encoder.MinHashEncoder.minhash",
    "dirty_cat.minhash_encoder.MinHashEncoder.set_params",
    "dirty_cat.minhash_encoder.MinHashEncoder.transform",
    "dirty_cat.similarity_encoder.SimilarityEncoder",
    "dirty_cat.similarity_encoder.SimilarityEncoder.fit",
    "dirty_cat.similarity_encoder.SimilarityEncoder.fit_transform",
    "dirty_cat.similarity_encoder.SimilarityEncoder.get_feature_names",
    "dirty_cat.similarity_encoder.SimilarityEncoder.get_feature_names_out",
    "dirty_cat.similarity_encoder.SimilarityEncoder.get_most_frequent",
    "dirty_cat.similarity_encoder.SimilarityEncoder.get_params",
    "dirty_cat.similarity_encoder.SimilarityEncoder.infrequent_categories_",
    "dirty_cat.similarity_encoder.SimilarityEncoder.inverse_transform",
    "dirty_cat.similarity_encoder.SimilarityEncoder.set_params",
    "dirty_cat.similarity_encoder.SimilarityEncoder.transform",
    "dirty_cat.super_vectorizer.SuperVectorizer",
    "dirty_cat.super_vectorizer.SuperVectorizer.OptionalEstimator",
    "dirty_cat.super_vectorizer.SuperVectorizer.fit",
    "dirty_cat.super_vectorizer.SuperVectorizer.fit_transform",
    "dirty_cat.super_vectorizer.SuperVectorizer.get_feature_names",
    "dirty_cat.super_vectorizer.SuperVectorizer.get_feature_names_out",
    "dirty_cat.super_vectorizer.SuperVectorizer.get_params",
    "dirty_cat.super_vectorizer.SuperVectorizer.named_transformers_",
    "dirty_cat.super_vectorizer.SuperVectorizer.set_params",
    "dirty_cat.super_vectorizer.SuperVectorizer.transform",
    "dirty_cat.target_encoder.TargetEncoder",
    "dirty_cat.target_encoder.TargetEncoder.fit",
    "dirty_cat.target_encoder.TargetEncoder.fit_transform",
    "dirty_cat.target_encoder.TargetEncoder.get_params",
    "dirty_cat.target_encoder.TargetEncoder.set_params",
    "dirty_cat.target_encoder.TargetEncoder.transform",
}


def all_estimators():
    module = import_module("dirty_cat")
    classes = inspect.getmembers(module, inspect.isclass)
    classes = [(name, est_cls) for name, est_cls in classes if not name.startswith("_")]
    return sorted(classes, key=lambda x: x[0])


def get_all_methods():
    estimators = all_estimators()
    for name, Estimator in estimators:
        if name.startswith("_"):
            # skip private classes
            continue
        methods = []
        for name in dir(Estimator):
            if name.startswith("_"):
                continue
            method_obj = getattr(Estimator, name)
            if hasattr(method_obj, "__call__") or isinstance(method_obj, property):
                methods.append(name)
        methods.append(None)

        for method in sorted(methods, key=str):
            yield Estimator, method


def repr_errors(res, estimator=None, method: Optional[str] = None) -> str:
    """Pretty print original docstring and the obtained errors

    Parameters
    ----------
    res : dict
        result of numpydoc.validate.validate
    estimator : {estimator, None}
        estimator object or None
    method : str
        if estimator is not None, either the method name or None.

    Returns
    -------
    str
       String representation of the error.
    """
    if method is None:
        if hasattr(estimator, "__init__"):
            method = "__init__"
        elif estimator is None:
            raise ValueError("At least one of estimator, method should be provided")
        else:
            raise NotImplementedError

    if estimator is not None:
        obj = getattr(estimator, method)
        try:
            obj_signature = str(inspect.signature(obj))
        except TypeError:
            # In particular we can't parse the signature of properties
            obj_signature = (
                "\nParsing of the method signature failed, "
                "possibly because this is a property."
            )

        obj_name = estimator.__name__ + "." + method
    else:
        obj_signature = ""
        obj_name = method

    msg = "\n\n" + "\n\n".join(
        [
            str(res["file"]),
            obj_name + obj_signature,
            res["docstring"],
            "# Errors",
            "\n".join(
                " - {}: {}".format(code, message) for code, message in res["errors"]
            ),
        ]
    )
    return msg


def filter_errors(errors, method, Estimator=None):
    """
    Ignore some errors based on the method type.
    """
    for code, message in errors:
        # We ignore following error code,
        #  - RT02: The first line of the Returns section
        #    should contain only the type, ..
        #   (as we may need refer to the name of the returned
        #    object)
        #  - GL01: Docstring text (summary) should start in the line
        #    immediately after the opening quotes (not in the same line,
        #    or leaving a blank line in between)
        #  - GL02: If there's a blank line, it should be before the
        #    first line of the Returns section, not after (it allows to have
        #    short docstrings for properties).

        if code in ["RT02", "GL01", "GL02"]:
            continue

        # Following codes are only taken into account for the
        # top level class docstrings:
        #  - ES01: No extended summary found
        #  - SA01: See Also section not found
        #  - EX01: No examples section found

        if method is not None and code in ["EX01", "SA01", "ES01"]:
            continue

        yield code, message


@pytest.mark.parametrize("Estimator, method", get_all_methods())
def test_docstring(Estimator, method, request):
    base_import_path = Estimator.__module__
    import_path = [base_import_path, Estimator.__name__]
    if method is not None:
        import_path.append(method)

    import_path = ".".join(import_path)

    if import_path in FUNCTION_DOCSTRING_IGNORE_SET:
        request.applymarker(
            pytest.mark.xfail(run=False, reason="TODO pass numpydoc validation")
        )

    res = numpydoc_validation.validate(import_path)

    res["errors"] = list(filter_errors(res["errors"], method, Estimator=Estimator))

    if res["errors"]:
        msg = repr_errors(res, Estimator, method)

        raise ValueError(msg)


if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser(description="Validate docstrings with numpydoc.")
    parser.add_argument("import_path", help="Import path to validate")

    args = parser.parse_args()

    res = numpydoc_validation.validate(args.import_path)

    import_path_sections = args.import_path.split(".")

    if len(import_path_sections) >= 2 and re.match(
        r"(?:[A-Z][a-z]*)+", import_path_sections[-2]
    ):
        method = import_path_sections[-1]
    else:
        method = None

    if res["errors"]:
        msg = repr_errors(res, method=args.import_path)

        print(msg)
        sys.exit(1)
    else:
        print("All docstring checks passed for {}!".format(args.import_path))



from typing import Any, Tuple

import numpy as np
import pandas as pd
import pytest
import sklearn
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from dirty_cat import GapEncoder, SuperVectorizer
from dirty_cat.utils import Version


def check_same_transformers(expected_transformers: dict, actual_transformers: list):
    # Construct the dict from the actual transformers
    actual_transformers_dict = {name: cols for name, trans, cols in actual_transformers}
    assert actual_transformers_dict == expected_transformers


def type_equality(expected_type, actual_type):
    """
    Checks that the expected type is equal to the actual type,
    assuming object and str types are equivalent
    (considered as categorical by the SuperVectorizer).
    """
    if (isinstance(expected_type, object) or isinstance(expected_type, str)) and (
        isinstance(actual_type, object) or isinstance(actual_type, str)
    ):
        return True
    else:
        return expected_type == actual_type


def _get_clean_dataframe() -> pd.DataFrame:
    """
    Creates a simple DataFrame with various types of data,
    and without missing values.
    """
    return pd.DataFrame(
        {
            "int": pd.Series([15, 56, 63, 12, 44], dtype="int"),
            "float": pd.Series([5.2, 2.4, 6.2, 10.45, 9.0], dtype="float"),
            "str1": pd.Series(
                ["public", "private", "private", "private", "public"], dtype="string"
            ),
            "str2": pd.Series(
                ["officer", "manager", "lawyer", "chef", "teacher"], dtype="string"
            ),
            "cat1": pd.Series(["yes", "yes", "no", "yes", "no"], dtype="category"),
            "cat2": pd.Series(
                ["20K+", "40K+", "60K+", "30K+", "50K+"], dtype="category"
            ),
        }
    )


def _get_dirty_dataframe() -> pd.DataFrame:
    """
    Creates a simple DataFrame with some missing values.
    We'll use different types of missing values (np.nan, pd.NA, None)
    to test the robustness of the vectorizer.
    """
    return pd.DataFrame(
        {
            "int": pd.Series([15, 56, pd.NA, 12, 44], dtype="Int64"),
            "float": pd.Series([5.2, 2.4, 6.2, 10.45, np.nan], dtype="Float64"),
            "str1": pd.Series(
                ["public", np.nan, "private", "private", "public"], dtype="object"
            ),
            "str2": pd.Series(
                ["officer", "manager", None, "chef", "teacher"], dtype="object"
            ),
            "cat1": pd.Series([np.nan, "yes", "no", "yes", "no"], dtype="object"),
            "cat2": pd.Series(["20K+", "40K+", "60K+", "30K+", np.nan], dtype="object"),
        }
    )


def _get_numpy_array() -> np.array:
    return np.array(
        [
            ["15", "56", pd.NA, "12", ""],
            ["?", "2.4", "6.2", "10.45", np.nan],
            ["public", np.nan, "private", "private", pd.NA],
            ["officer", "manager", None, "chef", "teacher"],
            [np.nan, "yes", "no", "yes", "no"],
            ["20K+", "40K+", "60K+", "30K+", np.nan],
        ]
    ).T


def _get_list_of_lists() -> list:
    return _get_numpy_array().tolist()


def _get_datetimes_dataframe() -> pd.DataFrame:
    """
    Creates a DataFrame with various date formats,
    already converted or to be converted.
    """
    return pd.DataFrame(
        {
            "pd_datetime": [
                pd.Timestamp("2019-01-01"),
                pd.Timestamp("2019-01-02"),
                pd.Timestamp("2019-01-03"),
                pd.Timestamp("2019-01-04"),
                pd.Timestamp("2019-01-05"),
            ],
            "np_datetime": [
                np.datetime64("2018-01-01"),
                np.datetime64("2018-01-02"),
                np.datetime64("2018-01-03"),
                np.datetime64("2018-01-04"),
                np.datetime64("2018-01-05"),
            ],
            "dmy-": [
                "11-12-2029",
                "02-12-2012",
                "11-09-2012",
                "13-02-2000",
                "10-11-2001",
            ],
            # "mdy-": ['11-13-2013',
            #          '02-12-2012',
            #          '11-31-2012',
            #          '05-02-2000',
            #          '10-11-2001'],
            "ymd/": [
                "2014/12/31",
                "2001/11/23",
                "2005/02/12",
                "1997/11/01",
                "2011/05/05",
            ],
            "ymd/_hms:": [
                "2014/12/31 00:31:01",
                "2014/12/30 00:31:12",
                "2014/12/31 23:31:23",
                "2015/12/31 01:31:34",
                "2014/01/31 00:32:45",
            ],
        }
    )


def _test_possibilities(X):
    """
    Do a bunch of tests with the SuperVectorizer.
    We take some expected transformers results as argument. They're usually
    lists or dictionaries.
    """
    # Test with low cardinality and a StandardScaler for the numeric columns
    vectorizer_base = SuperVectorizer(
        cardinality_threshold=4,
        # we must have n_samples = 5 >= n_components
        high_card_cat_transformer=GapEncoder(n_components=2),
        numerical_transformer=StandardScaler(),
    )
    # Warning: order-dependant
    expected_transformers_df = {
        "numeric": ["int", "float"],
        "low_card_cat": ["str1", "cat1"],
        "high_card_cat": ["str2", "cat2"],
    }
    vectorizer_base.fit_transform(X)
    check_same_transformers(expected_transformers_df, vectorizer_base.transformers)

    # Test with higher cardinality threshold and no numeric transformer
    expected_transformers_2 = {
        "low_card_cat": ["str1", "str2", "cat1", "cat2"],
    }
    vectorizer_default = SuperVectorizer()  # Using default values
    vectorizer_default.fit_transform(X)
    check_same_transformers(expected_transformers_2, vectorizer_default.transformers)

    # Test with a numpy array
    arr = X.to_numpy()
    # Instead of the columns names, we'll have the column indices.
    expected_transformers_np_no_cast = {
        "low_card_cat": [2, 4],
        "high_card_cat": [3, 5],
        "numeric": [0, 1],
    }
    vectorizer_base.fit_transform(arr)
    check_same_transformers(
        expected_transformers_np_no_cast, vectorizer_base.transformers
    )

    # Test with pandas series
    expected_transformers_series = {
        "low_card_cat": ["cat1"],
    }
    vectorizer_base.fit_transform(X["cat1"])
    check_same_transformers(expected_transformers_series, vectorizer_base.transformers)

    # Test casting values
    vectorizer_cast = SuperVectorizer(
        cardinality_threshold=4,
        # we must have n_samples = 5 >= n_components
        high_card_cat_transformer=GapEncoder(n_components=2),
        numerical_transformer=StandardScaler(),
    )
    X_str = X.astype("object")
    # With pandas
    expected_transformers_plain = {
        "high_card_cat": ["str2", "cat2"],
        "low_card_cat": ["str1", "cat1"],
        "numeric": ["int", "float"],
    }
    vectorizer_cast.fit_transform(X_str)
    check_same_transformers(expected_transformers_plain, vectorizer_cast.transformers)
    # With numpy
    expected_transformers_np_cast = {
        "numeric": [0, 1],
        "low_card_cat": [2, 4],
        "high_card_cat": [3, 5],
    }
    vectorizer_cast.fit_transform(X_str.to_numpy())
    check_same_transformers(expected_transformers_np_cast, vectorizer_cast.transformers)


def test_with_clean_data():
    """
    Defines the expected returns of the vectorizer in different settings,
    and runs the tests with a clean dataset.
    """
    _test_possibilities(_get_clean_dataframe())


def test_with_dirty_data() -> None:
    """
    Defines the expected returns of the vectorizer in different settings,
    and runs the tests with a dataset containing missing values.
    """
    _test_possibilities(_get_dirty_dataframe())


def test_auto_cast() -> None:
    """
    Tests that the SuperVectorizer automatic type detection works as expected.
    """
    vectorizer = SuperVectorizer()

    # Test datetime detection
    X = _get_datetimes_dataframe()

    expected_types_datetimes = {
        "pd_datetime": "datetime64[ns]",
        "np_datetime": "datetime64[ns]",
        "dmy-": "datetime64[ns]",
        "ymd/": "datetime64[ns]",
        "ymd/_hms:": "datetime64[ns]",
    }
    X_trans = vectorizer._auto_cast(X)
    for col in X_trans.columns:
        assert expected_types_datetimes[col] == X_trans[col].dtype

    # Test other types detection

    expected_types_clean_dataframe = {
        "int": "int64",
        "float": "float64",
        "str1": "object",
        "str2": "object",
        "cat1": "object",
        "cat2": "object",
    }

    X = _get_clean_dataframe()
    X_trans = vectorizer._auto_cast(X)
    for col in X_trans.columns:
        assert type_equality(expected_types_clean_dataframe[col], X_trans[col].dtype)

    # Test that missing values don't prevent type detection
    expected_types_dirty_dataframe = {
        "int": "float64",  # int type doesn't support nans
        "float": "float64",
        "str1": "object",
        "str2": "object",
        "cat1": "object",
        "cat2": "object",
    }

    X = _get_dirty_dataframe()
    X_trans = vectorizer._auto_cast(X)
    for col in X_trans.columns:
        assert type_equality(expected_types_dirty_dataframe[col], X_trans[col].dtype)


def test_with_arrays():
    """
    Check that the SuperVectorizer works if we input
    a list of lists or a numpy array.
    """
    expected_transformers = {
        "numeric": [0, 1],
        "low_card_cat": [2, 4],
        "high_card_cat": [3, 5],
    }
    vectorizer = SuperVectorizer(
        cardinality_threshold=4,
        # we must have n_samples = 5 >= n_components
        high_card_cat_transformer=GapEncoder(n_components=2),
        numerical_transformer=StandardScaler(),
    )

    X = _get_numpy_array()
    vectorizer.fit_transform(X)
    check_same_transformers(expected_transformers, vectorizer.transformers)

    X = _get_list_of_lists()
    vectorizer.fit_transform(X)
    check_same_transformers(expected_transformers, vectorizer.transformers)


def test_get_feature_names_out() -> None:
    X = _get_clean_dataframe()

    vec_w_pass = SuperVectorizer(remainder="passthrough")
    vec_w_pass.fit(X)

    # In this test, order matters. If it doesn't, convert to set.
    expected_feature_names_pass = [
        "str1_public",
        "str2_chef",
        "str2_lawyer",
        "str2_manager",
        "str2_officer",
        "str2_teacher",
        "cat1_yes",
        "cat2_20K+",
        "cat2_30K+",
        "cat2_40K+",
        "cat2_50K+",
        "cat2_60K+",
        "int",
        "float",
    ]
    if Version(sklearn.__version__) < Version("1.0"):
        assert vec_w_pass.get_feature_names() == expected_feature_names_pass
    else:
        assert vec_w_pass.get_feature_names_out() == expected_feature_names_pass

    vec_w_drop = SuperVectorizer(remainder="drop")
    vec_w_drop.fit(X)

    # In this test, order matters. If it doesn't, convert to set.
    expected_feature_names_drop = [
        "str1_public",
        "str2_chef",
        "str2_lawyer",
        "str2_manager",
        "str2_officer",
        "str2_teacher",
        "cat1_yes",
        "cat2_20K+",
        "cat2_30K+",
        "cat2_40K+",
        "cat2_50K+",
        "cat2_60K+",
    ]
    if Version(sklearn.__version__) < Version("1.0"):
        assert vec_w_drop.get_feature_names() == expected_feature_names_drop
    else:
        assert vec_w_drop.get_feature_names_out() == expected_feature_names_drop


def test_fit() -> None:
    # Simply checks sklearn's `check_is_fitted` function raises an error if
    # the SuperVectorizer is instantiated but not fitted.
    # See GH#193
    sup_vec = SuperVectorizer()
    with pytest.raises(NotFittedError):
        assert check_is_fitted(sup_vec)


def test_transform() -> None:
    X = _get_clean_dataframe()
    sup_vec = SuperVectorizer()
    sup_vec.fit(X)
    s = [34, 5.5, "private", "manager", "yes", "60K+"]
    x = np.array(s).reshape(1, -1)
    x_trans = sup_vec.transform(x)
    assert x_trans.tolist() == [
        [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 34.0, 5.5]
    ]
    # To understand the list above:
    # print(dict(zip(sup_vec.get_feature_names_out(), x_trans.tolist()[0])))


def test_fit_transform_equiv() -> None:
    """
    We will test the equivalence between using `.fit_transform(X)`
    and `.fit(X).transform(X).`
    """
    for X in [
        _get_clean_dataframe(),
        _get_dirty_dataframe(),
    ]:
        enc1_x1 = SuperVectorizer().fit_transform(X)
        enc2_x1 = SuperVectorizer().fit(X).transform(X)

        assert np.allclose(enc1_x1, enc2_x1, rtol=0, atol=0, equal_nan=True)


def _is_equal(elements: Tuple[Any, Any]) -> bool:
    """
    Fixture for values that return false when compared with `==`.
    """
    elem1, elem2 = elements  # Unpack
    return pd.isna(elem1) and pd.isna(elem2) or elem1 == elem2


def test_passthrough():
    """
    Tests that when passed no encoders, the SuperVectorizer
    returns the dataset as-is.
    """

    X_dirty = _get_dirty_dataframe()
    X_clean = _get_clean_dataframe()

    sv = SuperVectorizer(
        low_card_cat_transformer="passthrough",
        high_card_cat_transformer="passthrough",
        datetime_transformer="passthrough",
        numerical_transformer="passthrough",
        impute_missing="skip",
        auto_cast=False,
    )

    X_enc_dirty = pd.DataFrame(
        sv.fit_transform(X_dirty), columns=sv.get_feature_names_out()
    )
    X_enc_clean = pd.DataFrame(
        sv.fit_transform(X_clean), columns=sv.get_feature_names_out()
    )
    # Reorder encoded arrays' columns (see SV's doc "Notes" section as to why)
    X_enc_dirty = X_enc_dirty[X_dirty.columns]
    X_enc_clean = X_enc_clean[X_clean.columns]

    dirty_flat_df = X_dirty.to_numpy().ravel().tolist()
    dirty_flat_trans_df = X_enc_dirty.to_numpy().ravel().tolist()
    assert all(map(_is_equal, zip(dirty_flat_df, dirty_flat_trans_df)))
    assert (X_clean.to_numpy() == X_enc_clean.to_numpy()).all()


if __name__ == "__main__":
    print("start test_super_vectorizer with clean df")
    test_with_clean_data()
    print("passed")

    print("start test_super_vectorizer with dirty df")
    test_with_dirty_data()
    print("passed")

    print("start test_auto_cast")
    test_auto_cast()
    print("passed")

    print("start test_with_arrays")
    test_with_arrays()
    print("test_with_arrays passed")

    print("start  test_get_feature_names_out")
    test_get_feature_names_out()
    print("passed")

    print("start test_fit")
    test_fit()
    print("passed")

    print("start fit_transform_equiv")
    test_fit_transform_equiv()
    print("passed")

    print("start test_passthrough")
    test_passthrough()
    print("passed")

    print("Done")



"""
Performs a GridSearch to find the best parameters for the SuperVectorizer
among a selection.
"""

import logging
import pandas as pd

from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import Pipeline

from dirty_cat import SuperVectorizer
from dirty_cat.datasets import fetch_open_payments, fetch_drug_directory, \
    fetch_road_safety, fetch_midwest_survey, fetch_medical_charge, \
    fetch_employee_salaries, fetch_traffic_violations

from pathlib import Path
from functools import wraps
from datetime import datetime
from typing import List, Tuple


def get_classification_datasets() -> List[Tuple[dict, str]]:
    return [
        (fetch_open_payments(), 'open_payments'),
        # (fetch_drug_directory(), 'drug_directory),
        (fetch_road_safety(), 'road_safety'),
        (fetch_midwest_survey(), 'midwest_survey'),
        (fetch_traffic_violations(), 'traffic_violations'),
    ]


def get_regression_datasets() -> List[Tuple[dict, str]]:
    return [
        (fetch_medical_charge(), 'medical_charge'),
        (fetch_employee_salaries(), 'employee_salaries'),
    ]


def get_dataset(info) -> Tuple[pd.DataFrame, pd.Series]:
    df = pd.read_csv(info['path'], **info['read_csv_kwargs'])
    y = df[info['y']]
    X = df.drop(info['y'], axis=1).astype(str)
    return X, y


def set_logging(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        logging_level = logging.DEBUG

        logger = logging.getLogger()
        logger.setLevel(logging_level)

        formatter = logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s')
        formatter.datefmt = '%m/%d/%Y %H:%M:%S'

        path = Path(__file__).parent / f'tuning_{str(datetime.now())[:10]}.log'

        fh = logging.FileHandler(filename=path, mode='w')
        fh.setLevel(logging_level)
        fh.setFormatter(formatter)

        # sh = logging.StreamHandler(sys.stdout)
        # sh.setLevel(logging_level)
        # sh.setFormatter(formatter)

        logger.addHandler(fh)
        # logger.addHandler(sh)

        return func(*args, **kwargs)
    return wrapper


@set_logging
def main():
    logging.info('Launching !')

    card_possibilities = [20, 30, 40, 50]
    n_comp_possibilities = [10, 30, 50]

    logging.debug('Creating pipelines')
    regression_pipeline = Pipeline([
        ('sv', SuperVectorizer()),
        ('estimator', RandomForestRegressor()),
    ])
    classification_pipeline = Pipeline([
        ('sv', SuperVectorizer()),
        ('estimator', RandomForestClassifier()),
    ])

    logging.debug(f'With cardinality possibilities: {card_possibilities} '
                  f'and n_components possibilities: {n_comp_possibilities}')
    for pipeline, datasets in zip(
        [
            regression_pipeline,
            classification_pipeline,
        ],
        [
            get_regression_datasets(),
            get_classification_datasets(),
        ]
    ):
        for info, name in datasets:
            X, y = get_dataset(info)
            if name != 'traffic_violations':
                continue

            csv_path = Path('.').resolve() / f'{name}_results.csv'
            if csv_path.exists():
                # If the results already exist, we'll skip to the next
                logging.debug(f'Skipping {name} as {csv_path!s} was found')
                continue

            logging.debug(f'Running search on {name}')
            grid = GridSearchCV(
                estimator=pipeline,
                param_grid={
                    'sv__cardinality_threshold': card_possibilities,
                    'sv__high_card_str_transformer__n_components': n_comp_possibilities,
                },
                n_jobs=30,
            )
            grid.fit(X, y)

            df = pd.DataFrame(grid.cv_results_)
            df.to_csv(csv_path)
            logging.info(f'Saved search results in {csv_path!s}')


if __name__ == '__main__':
    main()



"""
Online Gamma-Poisson factorization of string arrays.
The principle is as follows:
    1. Given an input string array X, we build its bag-of-n-grams
       representation V (n_samples, vocab_size).
    2. Instead of using the n-grams counts as encodings, we look for low-
       dimensional representations by modeling n-grams counts as linear
       combinations of topics V = HW, with W (n_topics, vocab_size) the topics
       and H (n_samples, n_topics) the associated activations.
    3. Assuming that n-grams counts follow a Poisson law, we fit H and W to
       maximize the likelihood of the data, with a Gamma prior for the
       activations H to induce sparsity.
    4. In practice, this is equivalent to a non-negative matrix factorization
       with the Kullback-Leibler divergence as loss, and a Gamma prior on H.
       We thus optimize H and W with the multiplicative update method.
"""

import warnings
from typing import Dict, Generator, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.random import RandomState
from scipy import sparse
from sklearn import __version__ as sklearn_version
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.cluster import KMeans
from sklearn.feature_extraction.text import CountVectorizer, HashingVectorizer
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_random_state, gen_batches
from sklearn.utils.extmath import row_norms, safe_sparse_dot
from sklearn.utils.fixes import _object_dtype_isnan

from dirty_cat.utils import Version

from .utils import check_input

if Version(sklearn_version) < Version("0.24"):
    from sklearn.cluster._kmeans import _k_init
else:
    from sklearn.cluster import kmeans_plusplus

from sklearn.decomposition._nmf import _beta_divergence


class GapEncoderColumn(BaseEstimator, TransformerMixin):

    """See GapEncoder's docstring."""

    rho_: float
    H_dict_: Dict[np.array, np.array]

    def __init__(
        self,
        n_components: int = 10,
        batch_size: int = 128,
        gamma_shape_prior: float = 1.1,
        gamma_scale_prior: float = 1.0,
        rho: float = 0.95,
        rescale_rho: bool = False,
        hashing: bool = False,
        hashing_n_features: int = 2**12,
        init: Literal["k-means++", "random", "k-means"] = "k-means++",
        tol: float = 1e-4,
        min_iter: int = 2,
        max_iter: int = 5,
        ngram_range: Tuple[int, int] = (2, 4),
        analyzer: Literal["word", "char", "char_wb"] = "char",
        add_words: bool = False,
        random_state: Optional[Union[int, RandomState]] = None,
        rescale_W: bool = True,
        max_iter_e_step: int = 20,
    ):
        self.ngram_range = ngram_range
        self.n_components = n_components
        self.gamma_shape_prior = gamma_shape_prior  # 'a' parameter
        self.gamma_scale_prior = gamma_scale_prior  # 'b' parameter
        self.rho = rho
        self.rescale_rho = rescale_rho
        self.batch_size = batch_size
        self.tol = tol
        self.hashing = hashing
        self.hashing_n_features = hashing_n_features
        self.max_iter = max_iter
        self.min_iter = min_iter
        self.init = init
        self.analyzer = analyzer
        self.add_words = add_words
        self.random_state = check_random_state(random_state)
        self.rescale_W = rescale_W
        self.max_iter_e_step = max_iter_e_step

    def _init_vars(self, X) -> Tuple[np.array, np.array, np.array]:
        """
        Build the bag-of-n-grams representation V of X and initialize
        the topics W.
        """
        # Init n-grams counts vectorizer
        if self.hashing:
            self.ngrams_count_ = HashingVectorizer(
                analyzer=self.analyzer,
                ngram_range=self.ngram_range,
                n_features=self.hashing_n_features,
                norm=None,
                alternate_sign=False,
            )
            if self.add_words:  # Init a word counts vectorizer if needed
                self.word_count_ = HashingVectorizer(
                    analyzer="word",
                    n_features=self.hashing_n_features,
                    norm=None,
                    alternate_sign=False,
                )
        else:
            self.ngrams_count_ = CountVectorizer(
                analyzer=self.analyzer, ngram_range=self.ngram_range, dtype=np.float64
            )
            if self.add_words:
                self.word_count_ = CountVectorizer(dtype=np.float64)

        # Init H_dict_ with empty dict to train from scratch
        self.H_dict_ = dict()
        # Build the n-grams counts matrix unq_V on unique elements of X
        unq_X, lookup = np.unique(X, return_inverse=True)
        unq_V = self.ngrams_count_.fit_transform(unq_X)
        if self.add_words:  # Add word counts to unq_V
            unq_V2 = self.word_count_.fit_transform(unq_X)
            unq_V = sparse.hstack((unq_V, unq_V2), format="csr")

        if not self.hashing:  # Build n-grams/word vocabulary
            if Version(sklearn_version) < Version("1.0"):
                self.vocabulary = self.ngrams_count_.get_feature_names()
            else:
                self.vocabulary = self.ngrams_count_.get_feature_names_out()
            if self.add_words:
                if Version(sklearn_version) < Version("1.0"):
                    self.vocabulary = np.concatenate(
                        (self.vocabulary, self.word_count_.get_feature_names())
                    )
                else:
                    self.vocabulary = np.concatenate(
                        (self.vocabulary, self.word_count_.get_feature_names_out())
                    )
        _, self.n_vocab = unq_V.shape
        # Init the topics W given the n-grams counts V
        self.W_, self.A_, self.B_ = self._init_w(unq_V[lookup], X)
        # Init the activations unq_H of each unique input string
        unq_H = _rescale_h(unq_V, np.ones((len(unq_X), self.n_components)))
        # Update self.H_dict_ with unique input strings and their activations
        self.H_dict_.update(zip(unq_X, unq_H))
        if self.rescale_rho:
            # Make update rate per iteration independent of the batch_size
            self.rho_ = self.rho ** (self.batch_size / len(X))
        return unq_X, unq_V, lookup

    def _get_H(self, X: np.array) -> np.array:
        """
        Return the bag-of-n-grams representation of X.
        """
        H_out = np.empty((len(X), self.n_components))
        for x, h_out in zip(X, H_out):
            h_out[:] = self.H_dict_[x]
        return H_out

    def _init_w(self, V: np.array, X) -> Tuple[np.array, np.array, np.array]:
        """
        Initialize the topics W.
        If self.init='k-means++', we use the init method of
        sklearn.cluster.KMeans.
        If self.init='random', topics are initialized with a Gamma
        distribution.
        If self.init='k-means', topics are initialized with a KMeans on the
        n-grams counts.
        """
        if self.init == "k-means++":
            if Version(sklearn_version) < Version("0.24"):
                W = (
                    _k_init(
                        V,
                        self.n_components,
                        x_squared_norms=row_norms(V, squared=True),
                        random_state=self.random_state,
                        n_local_trials=None,
                    )
                    + 0.1
                )
            else:
                W, _ = kmeans_plusplus(
                    V,
                    self.n_components,
                    x_squared_norms=row_norms(V, squared=True),
                    random_state=self.random_state,
                    n_local_trials=None,
                )
                W = W + 0.1  # To avoid restricting topics to a few n-grams only
        elif self.init == "random":
            W = self.random_state.gamma(
                shape=self.gamma_shape_prior,
                scale=self.gamma_scale_prior,
                size=(self.n_components, self.n_vocab),
            )
        elif self.init == "k-means":
            prototypes = get_kmeans_prototypes(
                X,
                self.n_components,
                analyzer=self.analyzer,
                random_state=self.random_state,
            )
            W = self.ngrams_count_.transform(prototypes).A + 0.1
            if self.add_words:
                W2 = self.word_count_.transform(prototypes).A + 0.1
                W = np.hstack((W, W2))
            # if k-means doesn't find the exact number of prototypes
            if W.shape[0] < self.n_components:
                if Version(sklearn_version) < Version("0.24"):
                    W2 = (
                        _k_init(
                            V,
                            self.n_components - W.shape[0],
                            x_squared_norms=row_norms(V, squared=True),
                            random_state=self.random_state,
                            n_local_trials=None,
                        )
                        + 0.1
                    )
                else:
                    W2, _ = kmeans_plusplus(
                        V,
                        self.n_components - W.shape[0],
                        x_squared_norms=row_norms(V, squared=True),
                        random_state=self.random_state,
                        n_local_trials=None,
                    )
                    W2 = W2 + 0.1
                W = np.concatenate((W, W2), axis=0)
        else:
            raise ValueError(f"Initialization method {self.init!r} does not exist. ")
        W /= W.sum(axis=1, keepdims=True)
        A = np.ones((self.n_components, self.n_vocab)) * 1e-10
        B = A.copy()
        return W, A, B

    def fit(self, X, y=None) -> "GapEncoderColumn":
        """
        Fit the GapEncoder on batches of X.

        Parameters
        ----------
        X : array-like, shape (n_samples, )
            The string data to fit the model on.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        self
            Fitting GapEncoderColumn instance.
        """
        # Copy parameter rho
        self.rho_ = self.rho
        # Check if first item has str or np.str_ type
        assert isinstance(X[0], str), "Input data is not string. "
        # Make n-grams counts matrix unq_V
        unq_X, unq_V, lookup = self._init_vars(X)
        n_batch = (len(X) - 1) // self.batch_size + 1
        del X
        # Get activations unq_H
        unq_H = self._get_H(unq_X)

        for n_iter_ in range(self.max_iter):
            # Loop over batches
            for i, (unq_idx, idx) in enumerate(batch_lookup(lookup, n=self.batch_size)):
                if i == n_batch - 1:
                    W_last = self.W_.copy()
                # Update activations unq_H
                unq_H[unq_idx] = _multiplicative_update_h(
                    unq_V[unq_idx],
                    self.W_,
                    unq_H[unq_idx],
                    epsilon=1e-3,
                    max_iter=self.max_iter_e_step,
                    rescale_W=self.rescale_W,
                    gamma_shape_prior=self.gamma_shape_prior,
                    gamma_scale_prior=self.gamma_scale_prior,
                )
                # Update the topics self.W_
                _multiplicative_update_w(
                    unq_V[idx],
                    self.W_,
                    self.A_,
                    self.B_,
                    unq_H[idx],
                    self.rescale_W,
                    self.rho_,
                )

                if i == n_batch - 1:
                    # Compute the norm of the update of W in the last batch
                    W_change = np.linalg.norm(self.W_ - W_last) / np.linalg.norm(W_last)

            if (W_change < self.tol) and (n_iter_ >= self.min_iter - 1):
                break  # Stop if the change in W is smaller than the tolerance

        # Update self.H_dict_ with the learned encoded vectors (activations)
        self.H_dict_.update(zip(unq_X, unq_H))
        return self

    def get_feature_names(self, n_labels=3, prefix=""):
        """
        Ensures compatibility with sklearn < 1.0.
        Use `get_feature_names_out` instead.
        """
        warnings.warn(
            "Following the changes in scikit-learn 1.0, "
            "get_feature_names is deprecated. "
            "Use get_feature_names_out instead. ",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_feature_names_out(n_labels=n_labels, prefix=prefix)

    def get_feature_names_out(
        self,
        n_labels: int = 3,
        prefix: str = "",
    ) -> List[str]:
        """
        Returns the labels that best summarize the learned components/topics.
        For each topic, labels with the highest activations are selected.

        Parameters
        ----------
        n_labels : int, default=3
            The number of labels used to describe each topic.
        prefix : str, default=""
            Used as a prefix for the categories.

        Returns
        -------
        topic_labels : typing.List[str]
            The labels that best describe each topic.
        """

        vectorizer = CountVectorizer()
        vectorizer.fit(list(self.H_dict_.keys()))
        if Version(sklearn_version) < Version("1.0"):
            vocabulary = np.array(vectorizer.get_feature_names())
        else:
            vocabulary = np.array(vectorizer.get_feature_names_out())
        encoding = self.transform(np.array(vocabulary).reshape(-1))
        encoding = abs(encoding)
        encoding = encoding / np.sum(encoding, axis=1, keepdims=True)
        n_components = encoding.shape[1]
        topic_labels = []
        for i in range(n_components):
            x = encoding[:, i]
            labels = vocabulary[np.argsort(-x)[:n_labels]]
            topic_labels.append(labels)
        topic_labels = [prefix + ", ".join(label) for label in topic_labels]
        return topic_labels

    def score(self, X) -> float:
        """
        Returns the Kullback-Leibler divergence between the n-grams counts
        matrix V of X, and its non-negative factorization HW.

        Parameters
        ----------
        X : array-like (str), shape (n_samples, )
            The data to encode.

        Returns
        -------
        float.
            The Kullback-Leibler divergence.
        """

        # Build n-grams/word counts matrix
        unq_X, lookup = np.unique(X, return_inverse=True)
        unq_V = self.ngrams_count_.transform(unq_X)
        if self.add_words:
            unq_V2 = self.word_count_.transform(unq_X)
            unq_V = sparse.hstack((unq_V, unq_V2), format="csr")

        self._add_unseen_keys_to_H_dict(unq_X)
        unq_H = self._get_H(unq_X)
        # Given the learnt topics W, optimize the activations H to fit V = HW
        for slice in gen_batches(n=unq_H.shape[0], batch_size=self.batch_size):
            unq_H[slice] = _multiplicative_update_h(
                unq_V[slice],
                self.W_,
                unq_H[slice],
                epsilon=1e-3,
                max_iter=self.max_iter_e_step,
                rescale_W=self.rescale_W,
                gamma_shape_prior=self.gamma_shape_prior,
                gamma_scale_prior=self.gamma_scale_prior,
            )
        # Compute the KL divergence between V and HW
        kl_divergence = _beta_divergence(
            unq_V[lookup], unq_H[lookup], self.W_, "kullback-leibler", square_root=False
        )
        return kl_divergence

    def partial_fit(self, X, y=None) -> "GapEncoderColumn":
        """
        Partial fit of the GapEncoder on X.
        To be used in an online learning procedure where batches of data are
        coming one by one.

        Parameters
        ----------
        X : array-like, shape (n_samples, )
            The string data to fit the model on.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        GapEncoderColumn
            The fitted GapEncoderColumn instance.
        """

        # Init H_dict_ with empty dict if it's the first call of partial_fit
        if not hasattr(self, "H_dict_"):
            self.H_dict_ = dict()
        # Same thing for the rho_ parameter
        if not hasattr(self, "rho_"):
            self.rho_ = self.rho
        # Check if first item has str or np.str_ type
        assert isinstance(X[0], str), "Input data is not string. "
        # Check if it is not the first batch
        if hasattr(self, "vocabulary"):  # Update unq_X, unq_V with new batch
            unq_X, lookup = np.unique(X, return_inverse=True)
            unq_V = self.ngrams_count_.transform(unq_X)
            if self.add_words:
                unq_V2 = self.word_count_.transform(unq_X)
                unq_V = sparse.hstack((unq_V, unq_V2), format="csr")

            unseen_X = np.setdiff1d(unq_X, np.array([*self.H_dict_]))
            unseen_V = self.ngrams_count_.transform(unseen_X)
            if self.add_words:
                unseen_V2 = self.word_count_.transform(unseen_X)
                unseen_V = sparse.hstack((unseen_V, unseen_V2), format="csr")

            if unseen_V.shape[0] != 0:
                unseen_H = _rescale_h(
                    unseen_V, np.ones((len(unseen_X), self.n_components))
                )
                for x, h in zip(unseen_X, unseen_H):
                    self.H_dict_[x] = h
                del unseen_H
            del unseen_X, unseen_V
        else:  # If it is the first batch, call _init_vars to init unq_X, unq_V
            unq_X, unq_V, lookup = self._init_vars(X)

        unq_H = self._get_H(unq_X)
        # Update unq_H, the activations
        unq_H = _multiplicative_update_h(
            unq_V,
            self.W_,
            unq_H,
            epsilon=1e-3,
            max_iter=self.max_iter_e_step,
            rescale_W=self.rescale_W,
            gamma_shape_prior=self.gamma_shape_prior,
            gamma_scale_prior=self.gamma_scale_prior,
        )
        # Update the topics self.W_
        _multiplicative_update_w(
            unq_V[lookup],
            self.W_,
            self.A_,
            self.B_,
            unq_H[lookup],
            self.rescale_W,
            self.rho_,
        )
        # Update self.H_dict_ with the learned encoded vectors (activations)
        self.H_dict_.update(zip(unq_X, unq_H))
        return self

    def _add_unseen_keys_to_H_dict(self, X) -> None:
        """
        Add activations of unseen string categories from X to H_dict.
        """
        unseen_X = np.setdiff1d(X, np.array([*self.H_dict_]))
        if unseen_X.size > 0:
            unseen_V = self.ngrams_count_.transform(unseen_X)
            if self.add_words:
                unseen_V2 = self.word_count_.transform(unseen_X)
                unseen_V = sparse.hstack((unseen_V, unseen_V2), format="csr")

            unseen_H = _rescale_h(
                unseen_V, np.ones((unseen_V.shape[0], self.n_components))
            )
            self.H_dict_.update(zip(unseen_X, unseen_H))

    def transform(self, X) -> np.array:
        """
        Return the encoded vectors (activations) H of input strings in X.
        Given the learnt topics W, the activations H are tuned to fit V = HW.

        Parameters
        ----------
        X : array-like, shape (n_samples)
            The string data to encode.

        Returns
        -------
        H : 2-d array, shape (n_samples, n_topics)
            Transformed input.
        """
        # Check if first item has str or np.str_ type
        assert isinstance(X[0], str), "Input data is not string. "
        unq_X = np.unique(X)
        # Build the n-grams counts matrix V for the string data to encode
        unq_V = self.ngrams_count_.transform(unq_X)
        if self.add_words:  # Add words counts
            unq_V2 = self.word_count_.transform(unq_X)
            unq_V = sparse.hstack((unq_V, unq_V2), format="csr")
        # Add unseen strings in X to H_dict
        self._add_unseen_keys_to_H_dict(unq_X)
        unq_H = self._get_H(unq_X)
        # Loop over batches
        for slc in gen_batches(n=unq_H.shape[0], batch_size=self.batch_size):
            # Given the learnt topics W, optimize H to fit V = HW
            unq_H[slc] = _multiplicative_update_h(
                unq_V[slc],
                self.W_,
                unq_H[slc],
                epsilon=1e-3,
                max_iter=100,
                rescale_W=self.rescale_W,
                gamma_shape_prior=self.gamma_shape_prior,
                gamma_scale_prior=self.gamma_scale_prior,
            )
        # Store and return the encoded vectors of X
        self.H_dict_.update(zip(unq_X, unq_H))
        return self._get_H(X)


class GapEncoder(BaseEstimator, TransformerMixin):
    """
    This encoder can be understood as a continuous encoding on a set of latent
    categories estimated from the data. The latent categories are built by
    capturing combinations of substrings that frequently co-occur.

    The GapEncoder supports online learning on batches of data for
    scalability through the partial_fit method.

    Parameters
    ----------
    n_components : int, default=10
        Number of latent categories used to model string data.
    batch_size : int, default=128
        Number of samples per batch.
    gamma_shape_prior : float, default=1.1
        Shape parameter for the Gamma prior distribution.
    gamma_scale_prior : float, default=1.0
        Scale parameter for the Gamma prior distribution.
    rho : float, default=0.95
        Weight parameter for the update of the W matrix.
    rescale_rho : bool, default=False
        If true, use rho ** (batch_size / len(X)) instead of rho to obtain an
        update rate per iteration that is independent of the batch size.
    hashing : bool, default=False
        If true, HashingVectorizer is used instead of CountVectorizer.
        It has the advantage of being very low memory scalable to large
        datasets as there is no need to store a vocabulary dictionary in
        memory.
    hashing_n_features : int, default=2**12
        Number of features for the HashingVectorizer. Only relevant if
        hashing=True.
    init : typing.Literal["k-means++", "random", "k-means"], default='k-means++'
        Initialization method of the W matrix.
        Options: {'k-means++', 'random', 'k-means'}.
        If init='k-means++', we use the init method of sklearn.cluster.KMeans.
        If init='random', topics are initialized with a Gamma distribution.
        If init='k-means', topics are initialized with a KMeans on the n-grams
        counts. This usually makes convergence faster but is a bit slower.
    tol : float, default=1e-4
        Tolerance for the convergence of the matrix W.
    min_iter : int, default=2
        Minimum number of iterations on the input data.
    max_iter : int, default=5
        Maximum number of iterations on the input data.
    ngram_range : typing.Tuple[int, int], default=(2, 4)
        The range of ngram length that will be used to build the
        bag-of-n-grams representation of the input data.
    analyzer : typing.Literal["word", "char", "char_wb"], default='char'.
        Analyzer parameter for the CountVectorizer/HashingVectorizer.
        Options: {‘word’, ‘char’, ‘char_wb’}, describing whether the matrix V
        to factorize should be made of word counts or character n-gram counts.
        Option ‘char_wb’ creates character n-grams only from text inside word
        boundaries; n-grams at the edges of words are padded with space.
    add_words : bool, default=False
        If true, add the words counts to the bag-of-n-grams representation
        of the input data.
    random_state : typing.Optional[Union[int, RandomState]], default=None
        Pass an int for reproducible output across multiple function calls.
    rescale_W : bool, default=True
        If true, the weight matrix W is rescaled at each iteration
        to have a l1 norm equal to 1 for each row.
    max_iter_e_step : int, default=20
        Maximum number of iterations to adjust the activations h at each step.
    handle_missing : typing.Literal["error", "empty_impute"], default=empty_impute
        Whether to raise an error or impute with empty string '' if missing
        values (NaN) are present during fit (default is to impute).
        In the inverse transform, the missing category will be denoted as None.


    Attributes
    ----------
    rho_: float
    fitted_models_: typing.List[GapEncoderColumn]
    column_names_: typing.List[str]

    References
    ----------
    For a detailed description of the method, see
    `Encoding high-cardinality string categorical variables
    <https://hal.inria.fr/hal-02171256v4>`_ by Cerda, Varoquaux (2019).

    """

    rho_: float
    fitted_models_: List[GapEncoderColumn]
    column_names_: List[str]

    def __init__(
        self,
        n_components: int = 10,
        batch_size: int = 128,
        gamma_shape_prior: float = 1.1,
        gamma_scale_prior: float = 1.0,
        rho: float = 0.95,
        rescale_rho: bool = False,
        hashing: bool = False,
        hashing_n_features: int = 2**12,
        init: Literal["k-means++", "random", "k-means"] = "k-means++",
        tol: float = 1e-4,
        min_iter: int = 2,
        max_iter: int = 5,
        ngram_range: Tuple[int, int] = (2, 4),
        analyzer: Literal["word", "char", "char_wb"] = "char",
        add_words: bool = False,
        random_state: Optional[Union[int, RandomState]] = None,
        rescale_W: bool = True,
        max_iter_e_step: int = 20,
        handle_missing: Literal["error", "empty_impute"] = "zero_impute",
    ):
        self.ngram_range = ngram_range
        self.n_components = n_components
        self.gamma_shape_prior = gamma_shape_prior  # 'a' parameter
        self.gamma_scale_prior = gamma_scale_prior  # 'b' parameter
        self.rho = rho
        self.rescale_rho = rescale_rho
        self.batch_size = batch_size
        self.tol = tol
        self.hashing = hashing
        self.hashing_n_features = hashing_n_features
        self.max_iter = max_iter
        self.min_iter = min_iter
        self.init = init
        self.analyzer = analyzer
        self.add_words = add_words
        self.random_state = random_state
        self.rescale_W = rescale_W
        self.max_iter_e_step = max_iter_e_step
        self.handle_missing = handle_missing

    def _more_tags(self) -> Dict[str, List[str]]:
        """
        Used internally by sklearn to ease the estimator checks.
        """
        return {"X_types": ["categorical"]}

    def _create_column_gap_encoder(self) -> GapEncoderColumn:
        return GapEncoderColumn(
            ngram_range=self.ngram_range,
            n_components=self.n_components,
            analyzer=self.analyzer,
            gamma_shape_prior=self.gamma_shape_prior,
            gamma_scale_prior=self.gamma_scale_prior,
            rho=self.rho,
            rescale_rho=self.rescale_rho,
            batch_size=self.batch_size,
            tol=self.tol,
            hashing=self.hashing,
            hashing_n_features=self.hashing_n_features,
            max_iter=self.max_iter,
            init=self.init,
            add_words=self.add_words,
            random_state=self.random_state,
            rescale_W=self.rescale_W,
            max_iter_e_step=self.max_iter_e_step,
        )

    def _handle_missing(self, X):
        """
        Imputes missing values with `` or raises an error
        Note: modifies the array in-place.
        """
        if self.handle_missing not in ["error", "zero_impute"]:
            raise ValueError(
                "handle_missing should be either 'error' or "
                f"'zero_impute', got {self.handle_missing!r}. "
            )

        missing_mask = _object_dtype_isnan(X)

        if missing_mask.any():
            if self.handle_missing == "error":
                raise ValueError("Input data contains missing values. ")
            elif self.handle_missing == "zero_impute":
                X[missing_mask] = ""

        return X

    def fit(self, X, y=None) -> "GapEncoder":
        """
        Fit the GapEncoder on batches of X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The string data to fit the model on.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        GapEncoder
            Fitted GapEncoder instance.
        """

        # Copy parameter rho
        self.rho_ = self.rho
        # If X is a dataframe, store its column names
        if isinstance(X, pd.DataFrame):
            self.column_names_ = list(X.columns)
        # Check input data shape
        X = check_input(X)
        X = self._handle_missing(X)
        self.fitted_models_ = []
        for k in range(X.shape[1]):
            col_enc = self._create_column_gap_encoder()
            self.fitted_models_.append(col_enc.fit(X[:, k]))
        return self

    def transform(self, X) -> np.array:
        """
        Return the encoded vectors (activations) H of input strings in X.
        Given the learnt topics W, the activations H are tuned to fit V = HW.
        When X has several columns, they are encoded separately and
        then concatenated.

        Remark: calling transform multiple times in a row on the same
        input X can give slightly different encodings. This is expected
        due to a caching mechanism to speed things up.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The string data to encode.

        Returns
        -------
        H : 2-d array, shape (n_samples, n_topics * n_features)
            Transformed input.
        """

        # Check input data shape
        X = check_input(X)
        X = self._handle_missing(X)
        X_enc = []
        for k in range(X.shape[1]):
            X_enc.append(self.fitted_models_[k].transform(X[:, k]))
        X_enc = np.hstack(X_enc)
        return X_enc

    def partial_fit(self, X, y=None) -> "GapEncoder":
        """
        Partial fit of the GapEncoder on X.
        To be used in an online learning procedure where batches of data are
        coming one by one.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The string data to fit the model on.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        GapEncoder
            Fitted GapEncoder instance.
        """

        # If X is a dataframe, store its column names
        if isinstance(X, pd.DataFrame):
            self.column_names_ = list(X.columns)
        # Check input data shape
        X = check_input(X)
        X = self._handle_missing(X)
        # Init the `GapEncoderColumn` instances if the model was
        # not fitted already.
        if not hasattr(self, "fitted_models_"):
            self.fitted_models_ = [
                self._create_column_gap_encoder() for _ in range(X.shape[1])
            ]
        for k in range(X.shape[1]):
            self.fitted_models_[k].partial_fit(X[:, k])
        return self

    def get_feature_names_out(
        self,
        col_names: Optional[Union[Literal["auto"], List[str]]] = None,
        n_labels: int = 3,
    ):
        """
        Returns the labels that best summarize the learned components/topics.
        For each topic, labels with the highest activations are selected.

        Parameters
        ----------
        col_names : typing.Optional[typing.Union[typing.Literal["auto"], typing.List[str]]], default=None  # noqa
            The column names to be added as prefixes before the labels.
            If col_names == None, no prefixes are used.
            If col_names == 'auto', column names are automatically defined:
                - if the input data was a dataframe, its column names are used,
                - otherwise, 'col1', ..., 'colN' are used as prefixes.
            Prefixes can be manually set by passing a list for col_names.

        n_labels : int, default=3
            The number of labels used to describe each topic.

        Returns
        -------
        topic_labels : list of strings
            The labels that best describe each topic.
        """
        assert hasattr(
            self, "fitted_models_"
        ), "ERROR: GapEncoder must be fitted first."
        # Generate prefixes
        if isinstance(col_names, str) and col_names == "auto":
            if hasattr(self, "column_names_"):  # Use column names
                prefixes = ["%s: " % col for col in self.column_names_]
            else:  # Use 'col1: ', ... 'colN: ' as prefixes
                prefixes = ["col%d: " % i for i in range(len(self.fitted_models_))]
        elif col_names is None:  # Empty prefixes
            prefixes = [""] * len(self.fitted_models_)
        else:
            prefixes = ["%s: " % col for col in col_names]
        labels = list()
        for k, enc in enumerate(self.fitted_models_):
            col_labels = enc.get_feature_names_out(n_labels, prefixes[k])
            labels.extend(col_labels)
        return labels

    def get_feature_names(
        self, input_features=None, col_names: List[str] = None, n_labels: int = 3
    ) -> List[str]:
        """
        Ensures compatibility with sklearn < 1.0.
        Use `get_feature_names_out` instead.
        """
        if Version(sklearn_version) >= "1.0":
            warnings.warn(
                "Following the changes in scikit-learn 1.0, "
                "get_feature_names is deprecated. "
                "Use get_feature_names_out instead. ",
                DeprecationWarning,
                stacklevel=2,
            )
        return self.get_feature_names_out(col_names, n_labels)

    def score(self, X) -> float:
        """
        Returns the sum over the columns of X of the Kullback-Leibler
        divergence between the n-grams counts matrix V of X, and its
        non-negative factorization HW.

        Parameters
        ----------
        X : array-like (str), shape (n_samples, n_features)
            The data to encode.

        Returns
        -------
        float.
            The Kullback-Leibler divergence.
        """
        X = check_input(X)
        kl_divergence = 0
        for k in range(X.shape[1]):
            kl_divergence += self.fitted_models_[k].score(X[:, k])
        return kl_divergence


def _rescale_W(W: np.array, A: np.array) -> None:
    """
    Rescale the topics W to have a L1-norm equal to 1.
    Note that they are modified in-place.
    """
    s = W.sum(axis=1, keepdims=True)
    W /= s
    A /= s


def _multiplicative_update_w(
    Vt: np.array,
    W: np.array,
    A: np.array,
    B: np.array,
    Ht: np.array,
    rescale_W: bool,
    rho: float,
) -> Tuple[np.array, np.array, np.array]:
    """
    Multiplicative update step for the topics W.
    """
    A *= rho
    A += W * safe_sparse_dot(Ht.T, Vt.multiply(1 / (np.dot(Ht, W) + 1e-10)))
    B *= rho
    B += Ht.sum(axis=0).reshape(-1, 1)
    np.divide(A, B, out=W)
    if rescale_W:
        _rescale_W(W, A)
    return W, A, B


def _rescale_h(V: np.array, H: np.array) -> np.array:
    """
    Rescale the activations H.
    """
    epsilon = 1e-10  # in case of a document having length=0
    H *= np.maximum(epsilon, V.sum(axis=1).A)
    H /= H.sum(axis=1, keepdims=True)
    return H


def _multiplicative_update_h(
    Vt: np.array,
    W: np.array,
    Ht: np.array,
    epsilon: float = 1e-3,
    max_iter: int = 10,
    rescale_W: bool = False,
    gamma_shape_prior: float = 1.1,
    gamma_scale_prior: float = 1.0,
):
    """
    Multiplicative update step for the activations H.
    """
    if rescale_W:
        WT1 = 1 + 1 / gamma_scale_prior
        W_WT1 = W / WT1
    else:
        WT1 = np.sum(W, axis=1) + 1 / gamma_scale_prior
        W_WT1 = W / WT1.reshape(-1, 1)
    const = (gamma_shape_prior - 1) / WT1
    squared_epsilon = epsilon**2
    for vt, ht in zip(Vt, Ht):
        vt_ = vt.data
        idx = vt.indices
        W_WT1_ = W_WT1[:, idx]
        W_ = W[:, idx]
        squared_norm = 1
        for n_iter_ in range(max_iter):
            if squared_norm <= squared_epsilon:
                break
            aux = np.dot(W_WT1_, vt_ / (np.dot(ht, W_) + 1e-10))
            ht_out = ht * aux + const
            squared_norm = np.dot(ht_out - ht, ht_out - ht) / np.dot(ht, ht)
            ht[:] = ht_out
    return Ht


def batch_lookup(
    lookup: np.array,
    n: int = 1,
) -> Generator[Tuple[np.array, np.array], None, None]:
    """
    Make batches of the lookup array.
    """
    len_iter = len(lookup)
    for idx in range(0, len_iter, n):
        indices = lookup[slice(idx, min(idx + n, len_iter))]
        unq_indices = np.unique(indices)
        yield unq_indices, indices


def get_kmeans_prototypes(
    X,
    n_prototypes: int,
    analyzer: Literal["word", "char", "char_wb"] = "char",
    hashing_dim: int = 128,
    ngram_range: Tuple[int, int] = (2, 4),
    sparse: bool = False,
    sample_weight=None,
    random_state: Optional[Union[int, RandomState]] = None,
):
    """
    Computes prototypes based on:
      - dimensionality reduction (via hashing n-grams)
      - k-means clustering
      - nearest neighbor
    """
    vectorizer = HashingVectorizer(
        analyzer=analyzer,
        norm=None,
        alternate_sign=False,
        ngram_range=ngram_range,
        n_features=hashing_dim,
    )
    projected = vectorizer.transform(X)
    if not sparse:
        projected = projected.toarray()
    kmeans = KMeans(n_clusters=n_prototypes, random_state=random_state)
    kmeans.fit(projected, sample_weight=sample_weight)
    centers = kmeans.cluster_centers_
    neighbors = NearestNeighbors()
    neighbors.fit(projected)
    indexes_prototypes = np.unique(neighbors.kneighbors(centers, 1)[-1])
    return np.sort(X[indexes_prototypes])



from typing import Dict, List, Literal, Optional
from warnings import warn

import numpy as np
import pandas as pd
from sklearn import __version__ as sklearn_version
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from dirty_cat.utils import Version, check_input

WORD_TO_ALIAS: Dict[str, str] = {
    "year": "Y",
    "month": "M",
    "day": "D",
    "hour": "H",
    "minute": "min",
    "second": "S",
    "millisecond": "ms",
    "microsecond": "us",
    "nanosecond": "N",
}
TIME_LEVELS: List[str] = list(WORD_TO_ALIAS.keys())
AcceptedTimeValues = Literal[
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "nanosecond",
]


class DatetimeEncoder(BaseEstimator, TransformerMixin):
    """
    This encoder transforms each datetime column into several numeric columns
    corresponding to temporal features, e.g year, month, day...
    Constant extracted features are dropped; for instance, if the year is
    always the same in a feature, the extracted "year" column won't be added.
    If the dates are timezone aware, all the features extracted will correspond
    to the provided timezone.

    Parameters
    ----------
    extract_until : AcceptedTimeValues, default="hour"
        Extract up to this granularity.
        If all features have not been extracted, add the "total_time" feature,
        which contains the time to epoch (in seconds).
        For instance, if you specify "day", only "year", "month", "day" and
        "total_time" features will be created.
    add_day_of_the_week: bool, default=False
        Add day of the week feature (if day is extracted).
        This is a numerical feature from 0 (Monday) to 6 (Sunday).

    Attributes
    ----------
    n_features_in_: int
        Number of features in the data seen during fit.
    n_features_out_: int
        Number of features of the transformed data.
    features_per_column_: typing.Dict[int, typing.List[str]]
        Dictionary mapping the index of the original columns
        to the list of features extracted for each column.
    col_names_: typing.Optional[typing.List[str]]
        List of the names of the features of the input data,
        if input data was a pandas DataFrame, otherwise None.
    """

    n_features_in_: int
    n_features_out_: int
    features_per_column_: Dict[int, List[str]]
    col_names_: Optional[List[str]]

    def __init__(
        self,
        extract_until: AcceptedTimeValues = "hour",
        add_day_of_the_week: bool = False,
    ):
        self.extract_until = extract_until
        self.add_day_of_the_week = add_day_of_the_week

    def _more_tags(self):
        """
        Used internally by sklearn to ease the estimator checks.
        """
        return {"X_types": ["categorical"]}

    def _validate_keywords(self):
        if self.extract_until not in TIME_LEVELS:
            raise ValueError(
                f'"extract_until" should be one of {TIME_LEVELS}, '
                f"got {self.extract_until}. "
            )

    @staticmethod
    def _extract_from_date(date_series: pd.Series, feature: str):
        if feature == "year":
            return pd.DatetimeIndex(date_series).year.to_numpy()
        elif feature == "month":
            return pd.DatetimeIndex(date_series).month.to_numpy()
        elif feature == "day":
            return pd.DatetimeIndex(date_series).day.to_numpy()
        elif feature == "hour":
            return pd.DatetimeIndex(date_series).hour.to_numpy()
        elif feature == "minute":
            return pd.DatetimeIndex(date_series).minute.to_numpy()
        elif feature == "second":
            return pd.DatetimeIndex(date_series).second.to_numpy()
        elif feature == "millisecond":
            return pd.DatetimeIndex(date_series).millisecond.to_numpy()
        elif feature == "microsecond":
            return pd.DatetimeIndex(date_series).microsecond.to_numpy()
        elif feature == "nanosecond":
            return pd.DatetimeIndex(date_series).nanosecond.to_numpy()
        elif feature == "dayofweek":
            return pd.DatetimeIndex(date_series).dayofweek.to_numpy()
        elif feature == "total_time":
            tz = pd.DatetimeIndex(date_series).tz
            # Compute the time in seconds from the epoch time UTC
            if tz is None:
                return (
                    pd.to_datetime(date_series) - pd.Timestamp("1970-01-01")
                ) // pd.Timedelta("1s")
            else:
                return (
                    pd.DatetimeIndex(date_series).tz_convert("utc")
                    - pd.Timestamp("1970-01-01", tz="utc")
                ) // pd.Timedelta("1s")

    def fit(self, X, y=None) -> "DatetimeEncoder":
        """
        Fit the DatetimeEncoder to X. In practice, just stores which extracted
        features are not constant.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data where each column is a datetime feature.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        DatetimeEncoder
            Fitted DatetimeEncoder instance.

        """
        self._validate_keywords()
        # Columns to extract for each column,
        # before taking into account constant columns
        self._to_extract = TIME_LEVELS[: TIME_LEVELS.index(self.extract_until) + 1]
        if self.add_day_of_the_week:
            self._to_extract.append("dayofweek")
        if isinstance(X, pd.DataFrame):
            self.col_names_ = X.columns.to_list()
        else:
            self.col_names_ = None
        X = check_input(X)
        # Features to extract for each column, after removing constant features
        self.features_per_column_ = {}
        for i in range(X.shape[1]):
            self.features_per_column_[i] = []
        # Check which columns are constant
        for i in range(X.shape[1]):
            for feature in self._to_extract:
                if np.nanstd(self._extract_from_date(X[:, i], feature)) > 0:
                    self.features_per_column_[i].append(feature)
            # If some date features have not been extracted, then add the
            # "total_time" feature, which contains the full time to epoch
            remainder = (
                pd.to_datetime(X[:, i])
                - pd.to_datetime(
                    pd.DatetimeIndex(X[:, i]).floor(WORD_TO_ALIAS[self.extract_until])
                )
            ).seconds.to_numpy()
            if np.nanstd(remainder) > 0:
                self.features_per_column_[i].append("total_time")

        self.n_features_in_ = X.shape[1]
        self.n_features_out_ = len(
            np.concatenate(list(self.features_per_column_.values()))
        )

        return self

    def transform(self, X, y=None) -> np.array:
        """
        Transform X by replacing each datetime column with
        corresponding numerical features.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The data to transform, where each column is a datetime feature.
        y : None
            Unused, only here for compatibility.

        Returns
        -------
        np.array, shape (n_samples, n_features_out_)
            Transformed input.
        """
        check_is_fitted(
            self,
            attributes=["n_features_in_", "n_features_out_", "features_per_column_"],
        )
        X = check_input(X)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                f"The number of features in the input data ({X.shape[1]}) "
                "does not match the number of features "
                f"seen during fit ({self.n_features_in_}). "
            )
        # Create a new array with the extracted features,
        # choosing only features that weren't constant during fit
        X_ = np.empty((X.shape[0], self.n_features_out_), dtype=np.float64)
        idx = 0
        for i in range(X.shape[1]):
            for j, feature in enumerate(self.features_per_column_[i]):
                X_[:, idx + j] = self._extract_from_date(X[:, i], feature)
            idx += len(self.features_per_column_[i])
        return X_

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Returns clean feature names with format "<column_name>_<new_feature>"
        if the original data has column names, otherwise with format
        "<column_index>_<new_feature>" where `<new_feature>` is one of
        ["year", "month", "day", "hour", "minute", "second", "millisecond",
        "microsecond", "nanosecond", "dayofweek"].
        """
        feature_names = []
        for i in self.features_per_column_.keys():
            prefix = str(i) if self.col_names_ is None else self.col_names_[i]
            for feature in self.features_per_column_[i]:
                feature_names.append(f"{prefix}_{feature}")
        return feature_names

    def get_feature_names(self, input_features=None) -> List[str]:
        """
        Ensures compatibility with sklearn < 1.0, and returns the output of
        get_feature_names_out.
        """
        if Version(sklearn_version) >= "1.0":
            warn(
                "Following the changes in scikit-learn 1.0, "
                "get_feature_names is deprecated. "
                "Use get_feature_names_out instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        return self.get_feature_names_out()



from .fetching import (
    fetch_drug_directory,
    fetch_employee_salaries,
    fetch_medical_charge,
    fetch_midwest_survey,
    fetch_open_payments,
    fetch_road_safety,
    fetch_traffic_violations,
    get_data_dir,
)

__all__ = [
    "get_data_dir",
    "fetch_medical_charge",
    "fetch_midwest_survey",
    "fetch_employee_salaries",
    "fetch_road_safety",
    "fetch_open_payments",
    "fetch_drug_directory",
    "fetch_traffic_violations",
]



"""
Gets the results output by `supervectorizer_tuning.py`
and shows some graphs.
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from pathlib import Path


here = Path(__file__).parent

expected_end = '_results.csv'

param_name = "cardinality_threshold"
for file in here.iterdir():
    if file.name.endswith(expected_end):
        df: pd.DataFrame = pd.read_csv(file)
        name = file.name[:-len(expected_end)]

        card_unq = np.unique(df['param_sv__cardinality_threshold'])
        comp_unq = np.unique(df['param_sv__high_card_str_transformer__n_components'])

        card_amnt = len(card_unq)
        comp_amnt = len(comp_unq)

        scores = df["mean_test_score"].to_numpy().reshape(card_amnt, comp_amnt)

        fig, ax = plt.subplots()
        im = ax.imshow(scores)

        # We want to show all ticks...
        ax.set_xticks(np.arange(comp_amnt))
        ax.set_yticks(np.arange(card_amnt))
        # ... and label them with the respective list entries
        ax.set_xticklabels([f'n_components={v}' for v in comp_unq])
        ax.set_yticklabels([f'cardinality={v}' for v in card_unq])

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(card_amnt):
            for j in range(comp_amnt):
                text = ax.text(j, i, round(scores[i, j], 3),
                               ha="center", va="center", color="w")

        ax.set_title(name)
        fig.tight_layout()
        plt.show()
